# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "26/10/2020"


from silx.gui import qt
from tomwer.core.process.edit.imagekeyeditor import IMAGE_KEYS
from typing import Union
from tomwer.core.scan.hdf5scan import HDF5TomoScan
from tomoscan.esrf.scan.hdf5scan import ImageKey as _ImageKey
import weakref
import functools
from collections.abc import Iterable
import logging

_logger = logging.getLogger(__name__)


class ImageKeyDialog(qt.QDialog):
    """
    Dialog used to tune the ImageKey parameter.
    """

    sigValidated = qt.Signal()
    """Signal send when the user validate the image keys"""

    def __init__(self, parent):
        qt.QDialog.__init__(self, parent)

        self.setLayout(qt.QVBoxLayout())
        self._widget = ImageKeyWindow(self)
        self.layout().addWidget(self._widget)

        types = qt.QDialogButtonBox.Ok | qt.QDialogButtonBox.Cancel
        self._buttons = qt.QDialogButtonBox(self)
        self._buttons.setStandardButtons(types)
        self.layout().addWidget(self._buttons)

        # signal / slot connection
        self._buttons.button(qt.QDialogButtonBox.Ok).released.connect(
            self._validateReleased
        )
        self._buttons.button(qt.QDialogButtonBox.Ok).setText("validate")
        self._buttons.button(qt.QDialogButtonBox.Cancel).released.connect(
            self._resetReleased
        )
        self._buttons.button(qt.QDialogButtonBox.Cancel).setText("reset")

    def _validateReleased(self, *args, **kwargs):
        self.sigValidated.emit()

    def _resetReleased(self, *args, **kwargs):
        self._widget.reset()

    def getModifications(self):
        """

        :return: list of indexes to be modified
        :rtype: list
        """
        return self._widget.getModifications()

    def setModifications(self, modifications):
        self._widget.setModifications(modifications)

    def setScan(self, scan):
        """

        :param HDF5TomoScan scan: scan to be edited
        """
        self._widget.setScan(scan)


class ImageKeyWindow(qt.QMainWindow):
    def __init__(self, parent):
        qt.QMainWindow.__init__(self, parent)
        self.setWindowFlags(qt.Qt.Widget)

        # dock widget editor
        self._editorControl = _ImageKeyEditor(self)
        self._editorDockWidget = qt.QDockWidget(self)
        self._editorDockWidget.setWidget(self._editorControl)
        self._editorDockWidget.setFeatures(qt.QDockWidget.DockWidgetMovable)
        self.addDockWidget(qt.Qt.TopDockWidgetArea, self._editorDockWidget)

        # main window
        self._mainWindow = _ImageKeyListFiltered(parent=self)
        self.setCentralWidget(self._mainWindow)

        # signal / slot connection
        self._editorControl.sigApply.connect(self._updateModifications)

    def setScan(self, scan) -> None:
        """

        :param HDF5TomoScan scan: scan to be edited
        """
        self._mainWindow.setScan(scan=scan)
        self._editorControl.setScan(scan=scan)

    def _updateModifications(self):
        modifications = self._editorControl.getModifications()
        to_ = self._editorControl.getUpgradeToImagKey()
        self._mainWindow.applyModifications(
            modifications=modifications, new_image_key=to_
        )

    def getModifications(self):
        """

        :return: list of indexes to be modified
        :rtype: list
        """
        return self._mainWindow.getModifications()

    def setModifications(self, modifications):
        self._mainWindow.clearModifcations()
        image_keys_set = set(modifications.values())
        image_keys_set = set(
            [_ImageKey.from_value(image_key) for image_key in image_keys_set]
        )
        for image_key_type in image_keys_set:
            self._mainWindow.applyModifications(
                modifications=modifications, new_image_key=image_key_type
            )

    def reset(self):
        self._mainWindow.reset()


class _ImageKeyEditor(qt.QDialog):
    sigApply = qt.Signal()
    """Signal emitted when the user request to apply the current modifications
    """

    def __init__(self, parent):
        qt.QDialog.__init__(self, parent)
        self.setWindowFlags(qt.Qt.Widget)
        self.setLayout(qt.QGridLayout())

        # scan name + entry
        self._scanLabel = qt.QLabel("", parent=self)
        self.layout().addWidget(qt.QLabel("scan id"), 0, 0, 1, 1)
        self.layout().addWidget(self._scanLabel, 0, 1, 1, 1)

        # mode selection
        self.layout().addWidget(qt.QLabel("selection mode"), 1, 0, 1, 1)
        self._selectionModeCB = qt.QComboBox(self)
        for mode in ("range", "list"):
            self._selectionModeCB.addItem(mode)
        self._selectionModeCB.setCurrentIndex(self._selectionModeCB.findText("list"))
        self.layout().addWidget(self._selectionModeCB, 1, 1, 1, 1)

        # image selection
        self._selectionWidget = _EditionSelection(parent=self)
        self.layout().addWidget(self._selectionWidget, 2, 0, 1, 3)

        # upgrade to
        self._upgradeToLabel = qt.QLabel("Upgrade to:", self)
        self._upgradeToLabel.setAlignment(qt.Qt.AlignRight | qt.Qt.AlignVCenter)
        self.layout().addWidget(self._upgradeToLabel, 2, 3, 1, 1)
        self._upgradeToCB = qt.QComboBox(self)
        for image_key in IMAGE_KEYS:
            self._upgradeToCB.addItem(image_key)
        self.layout().addWidget(self._upgradeToCB, 2, 4, 1, 1)

        types = qt.QDialogButtonBox.Ok
        self._buttons = qt.QDialogButtonBox(self)
        self._buttons.setStandardButtons(types)
        self._buttons.button(qt.QDialogButtonBox.Ok).setText(
            "update modifications to apply"
        )
        self.layout().addWidget(self._buttons, 5, 0, 1, 5)
        self._warningLabel = qt.QLabel(
            "!!! `update list` this will only modify key value on the following list. To apply modification you must `validate` those modifications",
            self,
        )
        self._warningLabel.setStyleSheet("font: italic; color: red")
        self._warningLabel.setAlignment(qt.Qt.AlignRight)
        self.layout().addWidget(self._warningLabel, 6, 0, 1, 5)

        # setup
        self._selectionWidget.setSelectionMode(self._selectionModeCB.currentText())

        # connect signal / slot
        self._selectionModeCB.currentTextChanged.connect(
            self._selectionWidget.setSelectionMode
        )
        self._buttons.button(qt.QDialogButtonBox.Ok).clicked.connect(self._apply)

    def getUpgradeToImagKey(self) -> str:
        """

        :return: selected image key as literal
        :rtype: str
        """
        return self._upgradeToCB.currentText()

    def getSelectionMode(self) -> str:
        """

        :return: frame indexes selection mode. Can be 'range' or 'list'
        """
        return self._selectionModeCB.currentText()

    def getModifications(self) -> list:
        """

        :return: list of indexes to be modified
        :rtype: list
        """
        if self.getSelectionMode() == "range":
            return self._selectionWidget.getSlice()
        elif self.getSelectionMode() == "list":
            return self._selectionWidget.getList()
        else:
            raise NotImplementedError("Not implemented")

    def _apply(self):
        self.sigApply.emit()

    def setScan(self, scan):
        self._selectionWidget.setScan(scan=scan)
        if scan is not None:
            self._scanLabel.setText(scan.get_identifier().to_str())
        else:
            self._scanLabel.setText("")


class _SliceSelection(qt.QWidget):
    """
    Widget used to define a set of frames from a range
    """

    def __init__(self, parent):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QHBoxLayout())

        # from
        self._fromLabel = qt.QLabel("from: ", self)
        self._fromLabel.setAlignment(qt.Qt.AlignRight | qt.Qt.AlignVCenter)
        self.layout().addWidget(self._fromLabel)
        self._fromQLE = qt.QSpinBox(self)
        self._fromQLE.setRange(0, 1000000)
        self.layout().addWidget(self._fromQLE)
        # to
        self._toLabel = qt.QLabel("to: ", self)
        self._toLabel.setAlignment(qt.Qt.AlignRight | qt.Qt.AlignVCenter)
        self.layout().addWidget(self._toLabel)
        self._toQLE = qt.QSpinBox(self)
        self._toQLE.setRange(0, 1000000)
        self.layout().addWidget(self._toQLE)
        # steps
        self._stepsLabel = qt.QLabel("steps: ", self)
        self._stepsLabel.setAlignment(qt.Qt.AlignRight | qt.Qt.AlignVCenter)
        self.layout().addWidget(self._stepsLabel)
        self._stepsQLE = qt.QSpinBox(self)
        self._stepsQLE.setRange(1, 1000000)
        self.layout().addWidget(self._stepsQLE)

    def getSlice(self) -> slice:
        """

        :return: slice defined
        :rtype: slice
        """
        return slice(self._fromQLE.value(), self._toQLE.value(), self._stepsQLE.value())


class _ListSelection(qt.QWidget):
    """
    Widget to define a list of frames
    """

    def __init__(self, parent):
        qt.QWidget.__init__(self, parent)
        self._scan = None
        self.setLayout(qt.QHBoxLayout())

        # list selection
        self._listQLE = qt.QLineEdit("", self)
        self._listQLE.setPlaceholderText("0, 8, -1, ...")
        self.layout().addWidget(self._listQLE)
        reg_exp = qt.QRegExp(r"(\d+)(,[ ]?\s*\d+)*")
        # list of int with optional space
        validator = qt.QRegExpValidator(reg_exp)
        self._listQLE.setValidator(validator)

        # select button
        self._selectButton = qt.QPushButton("select", self)
        self.layout().addWidget(self._selectButton)

        # signal / slot selection
        self._selectButton.released.connect(self._selectFrames)

    def _selectFrames(self):
        if self._scan is None or self._scan() is None:
            _logger.warning(
                "No scan set or has been moved. Unable to select slices from it."
            )
            return
        scan = self._scan()
        dialog = _FrameSelectionDialog(parent=self)
        dialog.setFrames(scan.frames)
        if dialog.exec_():
            frame_indexes = dialog.getSelectedFrameIndexes()
            return self.setList(frame_indexes)

    def getList(self) -> list:
        """

        :return: list of indexes selected
        :rtype: list
        """
        text = self._listQLE.text()
        text = text.replace(" ", "")
        if text == "":
            return []
        else:
            return [int(element) for element in text.split(",")]

    def setList(self, list_: Union[str, Iterable]):
        """

        :param Union[str,Iterable] list_:
        :return:
        """
        if isinstance(list_, str):
            txt = list_
        else:
            txt = ", ".join(str(elmt) for elmt in list_)
        self._listQLE.setText(txt)

    def setScan(self, scan):
        self._scan = weakref.ref(scan)


class _EditionSelection(qt.QWidget):
    def __init__(self, parent):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QVBoxLayout())

        # slice selection
        self._sliceSelection = _SliceSelection(parent=self)
        self.layout().addWidget(self._sliceSelection)

        self._listSelection = _ListSelection(parent=self)
        self.layout().addWidget(self._listSelection)

    def setSelectionMode(self, mode: str) -> None:
        """

        :param str mode: define indexes selection mode. Can be 'range' or
                         'list'
        """
        if mode == "slice":
            # for backward compatibility (user can have .ows with old name)
            mode = "range"
        assert mode in ("range", "list")
        self._sliceSelection.setVisible(mode == "range")
        self._listSelection.setVisible(mode == "list")

    def getList(self) -> list:
        return self._listSelection.getList()

    def getSlice(self) -> slice:
        return self._sliceSelection.getSlice()

    def setScan(self, scan):
        self._listSelection.setScan(scan)


class _ImageKeyComboBox(qt.QComboBox):
    def __init__(self, parent, image_key):
        qt.QComboBox.__init__(self, parent)
        for img_key_type in IMAGE_KEYS:
            self.addItem(img_key_type)
        self.setImageKey(image_key)

    def setImageKey(self, image_key):
        if image_key in _ImageKeyList._IMAGE_KEYS_INV:
            image_key = _ImageKeyList._IMAGE_KEYS_INV[image_key]
        idx = self.findText(image_key)
        assert idx >= 0, "image key not recognized"
        self.setCurrentIndex(idx)


class _ImageKeyList(qt.QTableWidget):
    sigKeyUpdated = qt.Signal(tuple)
    """Signal emitted when a key has been updated. tuple contains
    (frame index, ImageKey)"""

    COLUMNS = (
        "index",
        "current image type",
        "new image type",
    )

    _IMAGE_KEYS_INV = {v: k for k, v in IMAGE_KEYS.items()}

    def __init__(self, parent):
        qt.QTableWidget.__init__(self, parent)
        self._filterMode = None
        self.__frames_modifications = {}
        """contains new and old values"""
        self._imageKeyComboboxes = {}
        """Qt / python needs to keep a pointer on those"""

    def setFilterMode(self, mode: str) -> None:
        """
        Define the filter mode to use and update the QTableWidget

        :param str mode: filtering mode
        """
        self._filterMode = mode
        self.updateFiltering()

    def clear(self) -> None:
        self._imageKeyComboboxes.clear()
        qt.QTableWidget.clear(self)

    def setFrames(self, frames: Iterable, frames_new_keys: dict) -> None:
        """
        Update the QTableWidget for the provided frames

        :param Iterable frames: Iterable of tomoscan.esrf.scan.HDF5TomoScan.Frame
        :param dict frames_new_keys: dictionary with frame edited. key is frame
                                     index. Value is the new `image_key` value
        """
        self.clear()
        self.setRowCount(len(frames))
        self.setColumnCount(len(self.COLUMNS))
        self.setHorizontalHeaderLabels(self.COLUMNS)
        self.horizontalHeader().setSectionResizeMode(qt.QHeaderView.ResizeToContents)
        self.verticalHeader().hide()

        for i_frame, frame in enumerate(frames):
            # frame index
            index = frame.index
            _item = qt.QTableWidgetItem()
            _item.setText(str(index))
            _item.setFlags(qt.Qt.ItemIsEnabled | qt.Qt.ItemIsSelectable)
            _item.setData(qt.Qt.UserRole, weakref.ref(frame))
            self.setItem(i_frame, 0, _item)

            # current image key
            current_image_key = _ImageKey.from_value(frame.image_key)
            current_image_key = self._IMAGE_KEYS_INV[current_image_key]
            currentImgKeyItem = qt.QLabel(current_image_key, self)
            self.setCellWidget(i_frame, 1, currentImgKeyItem)

            # new image key
            if frame.index in frames_new_keys:
                new_image_key = frames_new_keys[frame.index]
            else:
                new_image_key = frame.image_key
            new_image_key = _ImageKey.from_value(new_image_key)
            new_image_key = self._IMAGE_KEYS_INV[new_image_key]
            newImgKeyItem = _ImageKeyComboBox(parent=self, image_key=new_image_key)
            self._imageKeyComboboxes[index] = newImgKeyItem
            callback = functools.partial(self.notify_modified, frame.index)
            newImgKeyItem.currentIndexChanged.connect(callback)
            self.setCellWidget(i_frame, 2, newImgKeyItem)
        self.updateFiltering()

    def clearFrameModifications(self) -> None:
        """
        undo modifications
        """
        for frame_index in self.__frames_modifications:
            item_index = self._getFrameRowIndex(frame_index)
            if item_index:
                frame_item = self.item(item_index, 0).data(qt.Qt.UserRole)
                if frame_item() is not None:
                    new_image_key = frame_item().image_key
                    new_image_key = self._IMAGE_KEYS_INV[new_image_key]
                    old = self._imageKeyComboboxes[frame_index].blockSignals(True)
                    self._imageKeyComboboxes[frame_index].setImageKey(new_image_key)
                    self._imageKeyComboboxes[frame_index].blockSignals(old)
        self.__frames_modifications = {}
        self.updateFiltering()

    def applyModifications(
        self, modifications: Iterable, new_image_key: _ImageKey
    ) -> None:
        """
        Update the current QTableWidget with provided modications
        :param modifications: iterable of frame indexes to be modify
        :param new_image_key: new `image_key` value
        """
        if isinstance(modifications, slice):
            sl = modifications
            my_list = list(range(sl.start, sl.stop, sl.step))
        else:
            my_list = modifications
        to_apply = {}
        for frame_index in my_list:
            to_apply[frame_index] = new_image_key
            if frame_index in self._imageKeyComboboxes:
                self._imageKeyComboboxes[frame_index].setImageKey(new_image_key)

        for _, value in to_apply.items():
            assert isinstance(value, _ImageKey), "values should be instance of" + str(
                _ImageKey
            )
        self.__frames_modifications.update(to_apply)
        self.updateFiltering()

    def updateFiltering(self) -> None:
        for i in range(self.rowCount()):
            frame = self.item(i, 0).data(qt.Qt.UserRole)()
            if not frame:
                continue
            if self._filterMode is None:
                return
            if self._filterMode == _ImageKeyListFiltered.ALL_IMG_FILTER:
                visible = True
            elif self._filterMode == _ImageKeyListFiltered.MODIFIED_IMG_FILTER:
                visible = frame.index in self.__frames_modifications
            else:
                visible = frame.image_key == IMAGE_KEYS[self._filterMode]
            if visible:
                self.showRow(i)
            else:
                self.hideRow(i)

    def notify_modified(self, frame_index) -> None:
        """
        Update list of modification and emit sigKeyUpdated

        :param int frame_index:
        """
        frame_row = self._getFrameRowIndex(frame_index)
        if frame_row is not None:
            new_image_key = self._imageKeyComboboxes[frame_index].currentText()
            new_image_key = IMAGE_KEYS[new_image_key]
            self.__frames_modifications[frame_index] = new_image_key
            self.sigKeyUpdated.emit((frame_index, new_image_key))

    def _getFrameRowIndex(self, frame_index):
        for i in range(self.rowCount()):
            frame = self.item(i, 0).data(qt.Qt.UserRole)()
            if frame and frame.index == frame_index:
                return i
        return None

    def getModifications(self):
        return self.__frames_modifications


class _ImageKeyListFiltered(qt.QWidget):
    MODIFIED_IMG_FILTER = "modified"
    ALL_IMG_FILTER = "all"

    sigFilterchanged = qt.Signal()
    """Signal emitted when the filter to apply change"""

    def __init__(self, parent):
        qt.QWidget.__init__(self)
        self.setLayout(qt.QGridLayout())
        """Expected frames type once modifications will be done"""

        # add filter
        self.layout().addWidget(qt.QLabel("image filter:"), 0, 0, 1, 1)
        self._filterCB = qt.QComboBox(self)
        for filter_name in IMAGE_KEYS:
            self._filterCB.addItem(filter_name)
        self._filterCB.addItem(_ImageKeyListFiltered.MODIFIED_IMG_FILTER)
        self._filterCB.addItem(_ImageKeyListFiltered.ALL_IMG_FILTER)
        self.layout().addWidget(self._filterCB)

        # add TableWidget
        self._imageKeyList = _ImageKeyList(self)
        self.layout().addWidget(self._imageKeyList)

        # set up
        self._filterCB.setCurrentIndex(
            self._filterCB.findText(_ImageKeyListFiltered.ALL_IMG_FILTER)
        )
        self._updateFilter()

        # connect signal / slot
        self._filterCB.currentIndexChanged.connect(self._updateFilter)

    def applyModifications(self, modifications: Union[slice, list], new_image_key):
        if new_image_key in IMAGE_KEYS:
            new_image_key = IMAGE_KEYS[new_image_key]
        self._imageKeyList.applyModifications(modifications, new_image_key)

    def reset(self):
        self._imageKeyList.clearFrameModifications()

    def getModifications(self):
        return self._imageKeyList.getModifications()

    def clearModifcations(self):
        self._imageKeyList.clearFrameModifications()

    def getFilerMode(self):
        return self._filterCB.currentText()

    def setScan(self, scan):
        if not isinstance(scan, HDF5TomoScan):
            raise TypeError("scan should be an instance of {}" "".format(HDF5TomoScan))
        self.__frames_modifications = {}
        self._scan = weakref.ref(scan)
        frames = scan.frames
        self._imageKeyList.setFrames(frames, self.__frames_modifications)

    def _updateFilter(self):
        self._imageKeyList.setFilterMode(mode=self.getFilerMode())
        self._imageKeyList.updateFiltering()


class _FrameSelectionDialog(qt.QDialog):
    def __init__(self, parent):
        qt.QDialog.__init__(self, parent)
        self.setLayout(qt.QVBoxLayout())

        # list widget
        self._listWidget = qt.QListWidget(parent=self)
        self._listWidget.setSelectionMode(qt.QAbstractItemView.ExtendedSelection)
        self.layout().addWidget(self._listWidget)
        # add buttons
        types = qt.QDialogButtonBox.Ok | qt.QDialogButtonBox.Cancel
        self._buttons = qt.QDialogButtonBox(self)
        self._buttons.setStandardButtons(types)
        self.layout().addWidget(self._buttons)

        # connect signal / slot
        self._buttons.button(qt.QDialogButtonBox.Ok).clicked.connect(self.accept)
        self._buttons.button(qt.QDialogButtonBox.Cancel).clicked.connect(self.reject)

    def setFrames(self, frames):
        item_texts = []
        for frame in frames:
            item_texts.append(
                "{index} ({image_key})".format(
                    index=frame.index, image_key=frame.image_key.value
                )
            )
        self._listWidget.addItems(item_texts)

    def getSelectedFrameIndexes(self):
        item_selected = self._listWidget.selectedItems()

        def get_index(my_str):
            return int(my_str.replace(" ", "").split("(")[0])

        return [get_index(item.text()) for item in item_selected]


class ImageKeyUpgraderList(qt.QTableWidget):
    """
    List defining image key upgrade operations
    so to upgrade all frames with a specific Image_key to another image_key
    """

    COLUMNS = ("operations", "remove")

    sigOperationsChanged = qt.Signal()
    """signal emit when the operations changed"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._operations = {}
        # operations. dict with image_key 'from' as key and image_key to as value.
        # this ensure we will not ask for contradictory operations (like upgrading one image_key to two different image_key)
        self._update()

    def getOperations(self) -> tuple:
        return self._operations

    def setOperations(self, operations: dict):
        self._operations.clear()
        for from_image_key, to_image_key in operations.items():
            self.addOperation(from_image_key=from_image_key, to_image_key=to_image_key)

    def _update(self):
        self.setColumnCount(len(self.COLUMNS))
        self.setRowCount(len(self._operations))
        self.setHorizontalHeaderLabels(self.COLUMNS)
        self.horizontalHeader().setSectionResizeMode(qt.QHeaderView.ResizeToContents)
        self.verticalHeader().hide()

        for i_operation, (from_image_key, to_image_key) in enumerate(
            self._operations.items()
        ):
            item_info = qt.QTableWidgetItem()
            item_info.setText(
                f"convert all `{_ImageKeyList._IMAGE_KEYS_INV[from_image_key]}` frames to `{_ImageKeyList._IMAGE_KEYS_INV[to_image_key]}`"
            )
            self.setItem(i_operation, 0, item_info)

            remove_button = qt.QPushButton(parent=self)
            style = qt.QApplication.style()
            stopIcon = style.standardIcon(qt.QStyle.SP_DialogCloseButton)
            remove_button.setIcon(stopIcon)
            remove_button.released.connect(
                functools.partial(
                    self.removeOperation,
                    from_image_key,
                    to_image_key,
                )
            )
            self.setCellWidget(i_operation, 1, remove_button)

    def addOperation(self, from_image_key: _ImageKey, to_image_key: _ImageKey):
        from_image_key = _ImageKey.from_value(from_image_key)
        to_image_key = _ImageKey.from_value(to_image_key)

        self._operations[from_image_key] = to_image_key
        self._update()
        self.sigOperationsChanged.emit()

    def removeOperation(self, from_image_key: _ImageKey, to_image_key: _ImageKey):
        from_image_key = _ImageKey.from_value(from_image_key)
        to_image_key = _ImageKey.from_value(to_image_key)
        if (
            from_image_key not in self._operations
            or self._operations[from_image_key] != to_image_key
        ):
            return
        self._operations.pop(from_image_key)
        self._update()
        self.sigOperationsChanged.emit()


class _AddImageKeyUpgradeOperation(qt.QGroupBox):
    sigOperationAdded = qt.Signal(dict)
    """Signal emit when a new operation is requested. Dict contains keys 'from_image_key' and 'to_image_key' to define the operation"""

    def __init__(self, title="add operation", *args, **kwargs):
        super().__init__(title=title, *args, **kwargs)
        self.setLayout(qt.QHBoxLayout())
        self.layout().addWidget(qt.QLabel("from"))
        self._fromCB = _ImageKeyComboBox(parent=self, image_key=_ImageKey.PROJECTION)
        self.layout().addWidget(self._fromCB)
        self.layout().addWidget(qt.QLabel("to"))
        self._toCB = _ImageKeyComboBox(parent=self, image_key=_ImageKey.DARK_FIELD)
        self.layout().addWidget(self._toCB)
        self._addButtonPB = qt.QPushButton("add")
        self.layout().addWidget(self._addButtonPB)

        # signal / slot connection
        self._addButtonPB.released.connect(self._addButtonCallback)

    def _addButtonCallback(self, *args, **kwargs):
        self.sigOperationAdded.emit(self._getCurrentOperation())

    def _getCurrentOperation(self):
        return {
            "from_image_key": IMAGE_KEYS[self._fromCB.currentText()].value,
            "to_image_key": IMAGE_KEYS[self._toCB.currentText()].value,
        }


class ImageKeyUpgraderWidget(qt.QWidget):
    """
    Main widget to define ImageKey upgrade operations.
    Will upgrade all image_key within a specific value to another value.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setLayout(qt.QVBoxLayout())

        self._addOperationWidget = _AddImageKeyUpgradeOperation(parent=self)
        self.layout().addWidget(self._addOperationWidget)

        self._operationList = ImageKeyUpgraderList(parent=self)
        self.layout().addWidget(self._operationList)

        # connect signal / slot
        self._addOperationWidget.sigOperationAdded.connect(self._addOperationCallback)
        # expose signal
        self.sigOperationsChanged = self._operationList.sigOperationsChanged

    def _addOperationCallback(self, operation_as_dict: dict):
        return self.addOperation(**operation_as_dict)

    # exposing some API

    def getOperations(self) -> tuple:
        return self._operationList.getOperations()

    def addOperation(self, from_image_key: _ImageKey, to_image_key: _ImageKey):
        self._operationList.addOperation(
            from_image_key=from_image_key, to_image_key=to_image_key
        )

    def removeOperation(self, from_image_key: _ImageKey, to_image_key: _ImageKey):
        self._operationList.removeOperation(
            from_image_key=from_image_key, to_image_key=to_image_key
        )

    def setOperations(self, operations: dict):
        return self._operationList.setOperations(operations=operations)



# coding: utf-8
###########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################

"""contain the SADeltaBetaProcess. Half automatic best delta / beta finder
"""

__authors__ = [
    "H.Payno",
]
__license__ = "MIT"
__date__ = "28/10/2021"


import os
import logging
import h5py
import numpy
from copy import copy
from typing import Optional, Union
from nabu.pipeline.config import get_default_nabu_config
from nabu.pipeline.fullfield.nabu_config import (
    nabu_config as nabu_fullfield_default_config,
)
from processview.core.manager import ProcessManager, DatasetState
from processview.core.superviseprocess import SuperviseProcess
from tomwer.core.process.reconstruction.nabu.nabucommon import (
    ResultSlurmRun,
    ResultsLocalRun,
    ResultsWithStd,
)
from tomwer.core.scan.scanfactory import ScanFactory
from .params import SADeltaBetaParams
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.utils import logconfig
from tomwer.core.process.task import Task
from tomwer.core.scan.hdf5scan import HDF5TomoScan
from tomoscan.io import HDF5File
from tomoscan.esrf.scan.utils import get_data
import tomwer.version
from tomwer.core.process.reconstruction.nabu.nabuslices import (
    run_single_slice_reconstruction,
)
from tomwer.core.process.reconstruction.nabu.nabucommon import ResultsRun
from tomwer.core.progress import Progress
from ..nabu import utils as nabu_utils
from tomwer.core.process.reconstruction.scores import compute_score
from tomwer.core.process.reconstruction.scores import ScoreMethod
from tomwer.core.process.reconstruction.scores import ComputedScore
from tomwer.core.process.reconstruction.scores import get_disk_mask_radius, apply_roi
from tomwer.core.process.reconstruction.nabu.nabuslices import SingleSliceRunner
from tomwer.core.volume.volumefactory import VolumeFactory
from tomwer.core.utils.scanutils import data_identifier_to_scan

_logger = logging.getLogger(__name__)


DEFAULT_RECONS_FOLDER = "sadeltabeta_results"


def one_slice_several_db(
    scan: TomwerScanBase,
    configuration: Union[dict, SADeltaBetaParams],
    process_id: Optional[int] = None,
) -> tuple:
    """
    Run a slice reconstruction using nabu per Center Of Rotation (cor) provided
    Then for each compute a score (quality) of the center of rotation

    :param TomwerScanBase scan:
    :param Union[dict,SADeltaBetaParams] configuration:
    :return: cor_reconstructions, outs, errs
             cor_reconstructions is a dictionary of cor as key and a tuple
             (url, score) as value
    :rtype: tuple
    """
    if isinstance(configuration, dict):
        configuration = SADeltaBetaParams.from_dict(configuration)
    elif not isinstance(configuration, SADeltaBetaParams):
        raise TypeError(
            "configuration should be a dictionary or an instance of SAAxisParams"
        )

    if scan.axis_params is None:
        from tomwer.core.process.reconstruction.axis import AxisRP

        scan.axis_params = AxisRP()

    configuration.check_configuration()
    slice_index = configuration.slice_indexes
    delta_beta_s = configuration.delta_beta_values
    output_dir = configuration.output_dir
    dry_run = configuration.dry_run
    cluster_config = configuration.cluster_config
    _logger.info(
        "launch reconstruction of slice {} and delta / beta: {}".format(
            slice_index, delta_beta_s
        )
    )
    if isinstance(slice_index, str):
        if not slice_index == "middle":
            raise ValueError("slice index {} not recognized".format(slice_index))
    elif not len(slice_index) == 1:
        raise ValueError("only manage one slice")
    else:
        slice_index = list(slice_index.values())[0]
    advancement = Progress("sa-delta-beta - slice {} of {}".format(slice_index, scan))

    config = configuration.nabu_params.copy()

    _logger.info("start reconstruction of {}".format(str(scan)))
    # if scan contains some center of position copy it to nabu
    if scan.axis_params is not None and scan.axis_params.relative_cor_value is not None:
        if "reconstruction" in config:
            # move the cor value to the nabu reference
            cor_nabu_ref = scan.axis_params.relative_cor_value + scan.dim_1 / 2.0
            config["reconstruction"]["rotation_axis_position"] = str(cor_nabu_ref)

    _logger.info("set nabu reconstruction parameters to {}".format(str(scan)))
    scan.nabu_recons_params = config

    db_reconstructions = {}
    # key is delta / beta, value is url
    stderrs = []
    stdouts = []
    all_succeed = True
    future_tomo_objs = {}
    # key is delta / beta, value is future
    if advancement is not None:
        advancement.setMaxAdvancement(len(delta_beta_s))
    for db in delta_beta_s:
        if "output" not in config:
            config["output"] = {}
        if output_dir is None:
            config["output"]["location"] = os.path.join(
                scan.path, DEFAULT_RECONS_FOLDER
            )
        else:
            config["output"]["location"] = output_dir
        # TODO: allow file format modifications
        config["output"]["file_format"] = "hdf5"
        if "phase" not in config:
            config["phase"] = {}
        config["phase"]["delta_beta"] = (
            db,
        )  # warning: at this tage delta_beta expects a list of value
        config["phase"]["method"] = "Paganin"
        res = run_single_slice_reconstruction(
            nabu_config=config,
            scan=scan,
            slice_index=slice_index,
            dry_run=dry_run,
            ask_sinogram_registration=False,
            ask_sinogram_load=False,
            cluster_config=cluster_config,
            add_to_latest_reconstructions=False,
            process_id=process_id,
        )
        if isinstance(res, ResultsRun):
            all_succeed = all_succeed and res.success
        if isinstance(res, ResultSlurmRun):
            future_tomo_objs[db] = res.future_slurm_jobs
        if isinstance(res, ResultsWithStd):
            if slice_index is not None:
                stderrs.append(res.std_err)
                stdouts.append(res.std_out)
        if isinstance(res, ResultsLocalRun) and len(res.results_urls) > 0:
            assert len(res.results_urls) == 1, "only one slice expected"
            db_reconstructions[db] = res.results_urls[0]

        if advancement is not None:
            advancement.increaseAdvancement(1)

    pag = False
    ctf = False
    if "phase" in config:
        phase_method = config["phase"].get("method", "").lower()
        if phase_method in ("pag", "paganin"):
            pag = True
        elif phase_method in ("ctf",):
            ctf = True

    # treat future.
    for db, future_tomo_obj_list in future_tomo_objs.items():
        assert (
            len(future_tomo_obj_list) == 1
        ), "only one future should be created for one slice / db couple"
        future = future_tomo_obj_list[0]
        future.result()
        if future.cancelled() or future.exception():
            continue
        file_prefix = SingleSliceRunner.get_file_basename_reconstruction(
            scan=scan,
            slice_index=slice_index,
            pag=pag,
            db=int(db) if db is not None else None,
            ctf=ctf,
        )
        # retrieve url
        volume_identifier = nabu_utils.get_recons_volume_identifier(
            file_prefix=file_prefix,
            location=config["output"]["location"],
            file_format=config.get("file_format", "hdf5"),
            scan=scan,
            slice_index=None,
            start_z=None,
            end_z=None,
            expects_single_slice=True,
        )

        assert len(volume_identifier) <= 1, "only one slice expected"
        if len(volume_identifier) == 1:
            db_reconstructions[db] = volume_identifier[0]
        else:
            _logger.warning(
                f"something went wrong with reconstruction of {db} from {str(scan)}"
            )

    class PostProcessing:
        def run(self):
            datasets = self.load_datasets()

            mask_disk_radius = get_disk_mask_radius(datasets)
            scores = {}
            rois = {}
            for db, (url, data) in datasets.items():
                if data is None:
                    score = None
                else:
                    assert data.ndim == 2
                    data_roi = apply_roi(data=data, radius=mask_disk_radius, url=url)
                    rois[db] = data_roi

                    # move data_roi to [0-1] range
                    #  preprocessing: get percentile 0 and 99 from image and
                    #  "clean" highest and lowest pixels from it
                    min_p, max_p = numpy.percentile(data_roi, (1, 99))
                    data_roi_int = data_roi[...]
                    data_roi_int[data_roi_int < min_p] = min_p
                    data_roi_int[data_roi_int > max_p] = max_p
                    data_roi_int = (data_roi_int - min_p) / (max_p - min_p)

                    score = ComputedScore(
                        tv=compute_score(data=data_roi_int, method=ScoreMethod.TV),
                        std=compute_score(data=data_roi_int, method=ScoreMethod.STD),
                    )
                scores[db] = (url, score)
            return scores, rois

        def load_datasets(self):
            datasets_ = {}
            for db, volume_identifier in db_reconstructions.items():
                slice_url = None
                # in case the try processing fails
                try:
                    volume = VolumeFactory.create_tomo_object_from_identifier(
                        volume_identifier
                    )
                    volumes_urls = tuple(volume.browse_data_urls())
                    if len(volumes_urls) > 1:
                        _logger.warning(
                            f"found a volume with mode that one url ({volumes_urls})"
                        )
                    slice_url = volumes_urls[0]
                    data = get_data(slice_url)
                except Exception as e:
                    _logger.error(
                        "Fail to compute a score for {}. Reason is {}"
                        "".format(volume_identifier, str(e))
                    )
                    datasets_[db] = (slice_url, None)
                else:
                    if data.ndim == 3:
                        if data.shape[0] == 1:
                            data = data.reshape(data.shape[1], data.shape[2])
                        elif data.shape[2] == 1:
                            data = data.reshape(data.shape[0], data.shape[1])
                        else:
                            raise ValueError(
                                "Data is expected to be 2D. Not {}".format(data.ndim)
                            )
                    elif data.ndim == 2:
                        pass
                    else:
                        raise ValueError(
                            "Data is expected to be 2D. Not {}".format(data.ndim)
                        )

                    datasets_[db] = (slice_url, data)
            return datasets_

    post_processing = PostProcessing()
    scores, rois = post_processing.run()
    return scores, stdouts, stderrs, rois


class SADeltaBetaProcess(
    Task, SuperviseProcess, input_names=("data",), output_names=("data",)
):
    """
    Main process to launch several reconstruction of a single slice with
    several Center Of Rotation (cor) values
    """

    def __init__(
        self,
        process_id=None,
        varinfo=None,
        inputs=None,
        node_id=None,
        node_attrs=None,
        execinfo=None,
    ):
        Task.__init__(
            self,
            varinfo=varinfo,
            inputs=inputs,
            node_id=node_id,
            node_attrs=node_attrs,
            execinfo=execinfo,
        )
        SuperviseProcess.__init__(self, process_id=process_id)
        self._dry_run = inputs.get("dry_run", False)
        self._dump_process = inputs.get("dump_process", True)
        self._dump_roi = inputs.get("dump_roi", False)
        self._sa_delta_beta_params = inputs.get("sa_delta_beta_params", None)
        if self._sa_delta_beta_params is not None:
            self.set_configuration(self._sa_delta_beta_params)
        self._std_outs = tuple()
        self._std_errs = tuple()

    @property
    def dump_roi(self):
        return self._dump_roi

    @dump_roi.setter
    def dump_roi(self, dump):
        self._dump_roi = dump

    @property
    def std_outs(self):
        return self._std_outs

    @property
    def std_errs(self):
        return self._std_errs

    def set_dry_run(self, dry_run):
        self._dry_run = dry_run

    @property
    def dry_run(self):
        return self._dry_run

    def set_configuration(self, configuration: dict) -> None:
        if isinstance(configuration, SADeltaBetaParams):
            self._settings = configuration.to_dict()
        elif isinstance(configuration, dict):
            self._settings = configuration
        else:
            raise TypeError(
                "configuration should be an instance of dict or " "SAAxisParams"
            )

    @staticmethod
    def autofocus(scan) -> Optional[float]:
        scores = scan.sa_delta_beta_params.scores
        if scores is None:
            return
        score_method = scan.sa_delta_beta_params.score_method
        best_db, best_score = None, 0
        for cor, (_, score_cls) in scores.items():
            if score_cls is None:  # if score calculation failed
                continue
            score = score_cls.get(score_method)
            if score is None:
                continue
            if score > best_score:
                best_db, best_score = cor, score
        scan.sa_delta_beta_params.autofocus = best_db
        scan.sa_delta_beta_params.value = best_db
        return best_db

    def run(self):
        scan = data_identifier_to_scan(self.inputs.data)
        if scan is None:
            self.outputs.data = scan
            return
        if isinstance(scan, TomwerScanBase):
            scan = scan
        elif isinstance(scan, dict):
            scan = ScanFactory.create_scan_object_frm_dict(scan)
        else:
            raise ValueError(
                "input type of {}: {} is not managed" "".format(scan, type(scan))
            )
        # insure scan contains some parameter regarding sa delta / beta
        if scan.sa_delta_beta_params is None:
            scan.sa_delta_beta_params = SADeltaBetaParams()

        # create dir if does not exists
        config = copy(self.get_configuration())
        params = SADeltaBetaParams.from_dict(config)
        if params.output_dir in (None, ""):
            params.output_dir = os.path.join(scan.path, DEFAULT_RECONS_FOLDER)
            if not os.path.exists(params.output_dir):
                os.makedirs(params.output_dir)
        db_res, self._std_outs, self._std_errs, rois = one_slice_several_db(
            scan=scan,
            configuration=params.to_dict(),
            process_id=self.process_id,
        )
        scan.sa_delta_beta_params.scores = db_res
        best_db = self.autofocus(scan=scan)
        # store nabu recons parameters to be used within the nabu volume for example.
        config = self.get_configuration()["nabu_params"]
        # beam shape is not directly used by nabu (uses ctf_geometry directly)
        config.get("phase", {}).pop("beam_shape", None)
        sc_config = get_default_nabu_config(nabu_fullfield_default_config)
        sc_config.update(config)
        if best_db is not None:
            sc_config["phase"]["delta_beta"] = (
                best_db,
            )  # warning: at this tage delta_beta expects a list of value
        scan.nabu_recons_params = sc_config

        # end processing
        self._process_end(scan=scan, db_res=db_res, score_rois=rois)
        self.outputs.data = scan

    def _process_end(self, scan, db_res, score_rois):
        assert isinstance(scan, TomwerScanBase)
        try:
            extra = {
                logconfig.DOC_TITLE: self._scheme_title,
                logconfig.SCAN_ID: str(scan),
            }
            slice_index = self.get_configuration().get("slice_index", None)

            if db_res is None:
                info = (
                    "fail to compute delta/beta scores of slice {} for scan {}."
                    "".format(slice_index, scan)
                )
                _logger.processFailed(info, extra=extra)
                ProcessManager().notify_dataset_state(
                    dataset=scan, process=self, state=DatasetState.FAILED, details=info
                )
            else:
                info = "delta/beta scores of slice {} for scan {} computed." "".format(
                    slice_index, scan
                )
                _logger.processSucceed(info, extra=extra)
                ProcessManager().notify_dataset_state(
                    dataset=scan,
                    process=self,
                    state=DatasetState.WAIT_USER_VALIDATION,
                    details=info,
                )
        except Exception as e:
            _logger.error(e)
        else:
            if self._dump_process:
                process_idx = SADeltaBetaProcess.process_to_tomwer_processes(
                    scan=scan,
                )
                if self.dump_roi and process_idx is not None:
                    self.dump_rois(
                        scan, score_rois=score_rois, process_index=process_idx
                    )

    @staticmethod
    def dump_rois(scan, score_rois: dict, process_index: int):
        if score_rois is None or len(score_rois) == 0:
            return
        if not isinstance(score_rois, dict):
            raise TypeError("score_rois is expected to be a dict")
        process_file = scan.process_file
        process_name = "tomwer_process_" + str(process_index)

        if scan.saaxis_params.scores in (None, {}):
            return

        def get_process_path():
            return "/".join((scan.entry or "entry", process_name))

        # save it to the file
        with Task._get_lock(process_file):
            # needs an extra lock for multiprocessing

            with HDF5File(process_file, mode="a") as h5f:
                nx_process = h5f.require_group(get_process_path())
                score_roi_grp = nx_process.require_group("score_roi")
                for db, roi in score_rois.items():
                    score_roi_grp[str(db)] = roi
                    score_roi_grp[str(db)].attrs["interpretation"] = "image"

    @staticmethod
    def program_name():
        """Name of the program used for this processing"""
        return "semi-automatic delta/beta finder"

    @staticmethod
    def program_version():
        """version of the program used for this processing"""
        return tomwer.version.version

    @staticmethod
    def definition():
        """definition of the process"""
        return "Semi automatic center of rotation / axis calculation"

    @staticmethod
    def process_to_tomwer_processes(scan):
        if scan.process_file is not None:
            entry = "entry"
            if isinstance(scan, HDF5TomoScan):
                entry = scan.entry

            db = None
            if scan.sa_delta_beta_params is not None:
                db = scan.sa_delta_beta_params.selected_delta_beta_value

            process_index = scan.pop_process_index()
            try:
                with scan.acquire_process_file_lock():
                    Task._register_process(
                        process_file=scan.process_file,
                        entry=entry,
                        results={"delta_beta": db if db is not None else "-"},
                        configuration=scan.sa_delta_beta_params.to_dict(),
                        process_index=process_index,
                        overwrite=True,
                        process=SADeltaBetaProcess,
                    )
                    SADeltaBetaProcess._extends_results(
                        scan=scan, entry=entry, process_index=process_index
                    )
            except Exception as e:
                _logger.warning(
                    f"Fail to register process of with index {process_index}. Reason is {e}"
                )

            return process_index

    @staticmethod
    def _extends_results(scan, entry, process_index):
        process_file = scan.process_file
        process_name = "tomwer_process_" + str(process_index)

        if scan.sa_delta_beta_params.scores in (None, {}):
            return

        def get_process_path():
            return "/".join((entry or "entry", process_name))

        # save it to the file
        with Task._get_lock(process_file):
            # needs an extra lock for multiprocessing

            with HDF5File(process_file, mode="a") as h5f:
                nx_process = h5f.require_group(get_process_path())
                if "NX_class" not in nx_process.attrs:
                    nx_process.attrs["NX_class"] = "NXprocess"

                results = nx_process.require_group("results")
                for cor, (url, score) in scan.sa_delta_beta_params.scores.items():
                    results_db = results.require_group(str(cor))
                    for method in ScoreMethod:
                        if method is ScoreMethod.TOMO_CONSISTENCY:
                            continue
                        results_db[method.value] = score.get(method)

                    link_path = os.path.relpath(
                        url.file_path(),
                        os.path.dirname(process_file),
                    )
                    results_db["reconstructed_slice"] = h5py.ExternalLink(
                        link_path, url.data_path()
                    )



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "02/06/2017"


import os


__LBSRAM_PATH = '/lbsram'

__DEST_PATH = ''

MOCK_LBSRAM = False

MAX_MEM_USED = 80
"Maximal percentage of used memory from which we should skip some processing"

MAKE_OAR_PYST2_PATH = '/data/id19/inhouse/OAR_UTILITIES/pyhst/make_oar_pyhst2'
"""path to the make_oar_pyhst2 path"""

JSON_RPC_HOST = None
"""you specify an host name. If None is provided then will use HOSTNAME"""

JSON_RPC_PORT = 4000


def mock_lsbram(b):
    assert type(b) is bool
    global MOCK_LBSRAM
    MOCK_LBSRAM = b


def isOnLbsram(scan=None):
    if MOCK_LBSRAM:
        return True
    elif scan is None:
        return os.path.isdir(get_lbsram_path())
    else:
        if isinstance(scan, str):
            return os.path.abspath(scan).startswith(__LBSRAM_PATH)
        else:
            return os.path.abspath(scan.path).startswith(__LBSRAM_PATH)


def get_lbsram_path():
    return __LBSRAM_PATH


def get_dest_path():
    return __DEST_PATH


def _set_lbsram_path(path):
    global __LBSRAM_PATH
    __LBSRAM_PATH = path


def _set_dest_path(path):
    global __DEST_PATH
    __DEST_PATH = path


class SlurmSettings:
    # Default slurm cluster configuration
    N_CORES_PER_TASK = 4
    """Number of CPU per worker"""

    N_TASKS = 1
    """Number of worker"""

    MEMORY_PER_WORKER = 50  # memory in GB
    """Amount of memory per worker"""

    PARTITION = "p9gpu"
    """Queue / partition to use"""

    DEFAULT_WALLTIME = "01:00:00"  # None if the default wall time

    N_GPUS_PER_WORKER = 1
    """number of gpu per worker"""

    PYTHON_VENV = "/scisoft/tomotools/activate.sh stable"
    """Python executable to take. Useful if compute nodes have a different environment from the front end.
    """

    PROJECT_NAME = "tomwer_{scan}_-_{process}_-_{info}"
    """Slurm cluster project name. `scan`, `process` and `info` will be format.
    """



# coding: utf-8
###########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################

"""contain utils for score process
"""

__authors__ = [
    "H.Payno",
]
__license__ = "MIT"
__date__ = "28/10/2021"

try:
    from nabu.pipeline.fullfield.reconstruction import (  # noqa F401
        FullFieldReconstructor,
    )
except (ImportError, OSError):
    try:
        from nabu.pipeline.fullfield.local_reconstruction import (  # noqa F401
            ChunkedReconstructor,
        )
    except (ImportError, OSError):
        # import of cufft library can bring an OSError if cuda not install
        has_nabu = False
    else:
        has_nabu = True
else:
    has_nabu = True
from typing import Iterable, Optional, Union

from processview.core.manager.manager import ProcessManager
from tomwer.core.process.reconstruction.nabu.target import Target
from tomwer.core.progress import Progress
from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.core.futureobject import FutureTomwerObject
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.utils.slurm import is_slurm_available
from tomwer.utils import docstring
from ..nabu import settings as nabu_settings
from tomwer.core.process.reconstruction.nabu.nabuslices import (
    SingleSliceRunner,
    generate_nabu_configfile,
    _NabuBaseReconstructor,
)
from nabu.pipeline.fullfield.nabu_config import (
    nabu_config as nabu_fullfield_default_config,
)
from .nabucommon import ResultSlurmRun, ResultsLocalRun, ResultsWithStd
from . import utils
import logging
import os

_logger = logging.getLogger(__name__)


def run_nabu_one_slice_several_config(
    scan: TomwerScanBase,
    nabu_configs: Union[list, tuple],
    cluster_config: Optional[dict],
    dry_run: bool,
    slice_index: Union[int, str],
    file_format: str,
    advancement: Optional[Progress] = None,
    process_id: Optional[int] = None,
) -> tuple:
    """
    Run several reconstruction of a specific slice.

    :param scan: dataset
    :param Iterable nabu_configs: set of nabu configurations to be run
    :param bool dry_run:
    :param int slice_index: slice index to reconstruct or "middle"
    :param Progress advancement: optional class to display advancement
    :param Optional[int] process_id: id of the process requesting this computation
    :param Optional[dict] cluster_config: cluster configuration if
    :return: success, recons_urls (list of output urls), tuple of outs, tuples of errs, dict future_scans (key is cor, value is future_scan)
    :rtype: tuple
    """
    if cluster_config in (None, {}):
        target = Target.LOCAL
    elif isinstance(cluster_config, dict):
        if not is_slurm_available():
            raise RuntimeError("Slurm computation requested but unvailable")
        target = Target.SLURM
    else:
        raise TypeError(
            f"cluster_config should be None or a dict not {type(cluster_config)}"
        )

    if process_id is not None:
        try:
            process_name = ProcessManager().get_process(process_id=process_id).name
        except KeyError:
            process_name = "unknow"
    else:
        process_name = ""

    reconstructor = _Reconstructor(
        scan=scan,
        nabu_configs=nabu_configs,
        advancement=advancement,
        slice_index=slice_index,
        target=target,
        dry_run=dry_run,
        file_format=file_format,
        cluster_config=cluster_config,
        process_name=process_name,
    )
    try:
        results = reconstructor.run()
    except TimeoutError as e:
        _logger.error(e)
        return None
    else:
        assert isinstance(
            results, dict
        ), "results should be a dictionary with cor as key and urls as value"
        success = True
        recons_urls = {}
        std_outs = []
        std_errs = []
        future_tomo_objs = {}
        for cor, res in results.items():
            success = success and res.success
            if isinstance(res, ResultsWithStd):
                std_outs.append(res.std_out)
                std_errs.append(res.std_err)
            if isinstance(res, ResultsLocalRun):
                recons_urls[cor] = res.results_urls
            if isinstance(res, ResultSlurmRun):
                future_tomo_obj = FutureTomwerObject(
                    tomo_obj=scan,
                    process_requester_id=process_id,
                    futures=res.future_slurm_jobs,
                )
                future_tomo_objs[cor] = future_tomo_obj
        return success, recons_urls, std_outs, std_errs, future_tomo_objs


class _Reconstructor(_NabuBaseReconstructor):
    def __init__(
        self,
        scan: TomwerScanBase,
        nabu_configs: Iterable,
        advancement: Optional[Progress],
        slice_index: Union[int, str],
        target: Target,
        dry_run: bool,
        file_format: str,
        cluster_config: Optional[dict],
        process_name: str,
    ) -> None:
        super().__init__(
            scan=scan,
            dry_run=dry_run,
            target=target,
            cluster_config=cluster_config,
            process_name=process_name,
        )
        if not isinstance(slice_index, (int, str)):
            raise TypeError(
                f"slice_index should be an int or a string not {type(slice_index)}"
            )

        self.advancement = advancement
        self.slice_index = slice_index
        self.nabu_configs = nabu_configs
        self.file_format = file_format

    @docstring(_NabuBaseReconstructor)
    def run(self) -> Iterable:
        if self.slice_index == "middle":
            if self.scan.dim_2 is not None:
                self.slice_index = self.scan.dim_2 // 2
            else:
                _logger.warning(
                    "scan.dim_2 returns None, unable to deduce middle " "pick 1024"
                )
                self.slice_index = 1024

        results = {}
        if self.advancement:
            self.advancement.setMaxAdvancement(len(self.nabu_configs))
        for cor, config in self.nabu_configs.items():
            config, conf_file = self.preprocess_config(config, cor)

            # add some tomwer metadata and save the configuration
            # note: for now the section is ignored by nabu but shouldn't stay that way
            with utils.TomwerInfo(config) as config_to_dump:
                generate_nabu_configfile(
                    conf_file,
                    nabu_fullfield_default_config,
                    config=config_to_dump,
                    options_level="advanced",
                )

            results[cor] = self._process_config(
                config_to_dump=config_to_dump,
                config_file=conf_file,
                file_format=self.file_format,
                start_z=None,
                end_z=None,
                info="nabu slice reconstruction",
                process_name=self.process_name,
            )
            # specific treatment for cor: rename output files
            if self.advancement:
                self.advancement.increaseAdvancement(1)
        return results

    def treateOutputConfig(self, _config, cor):
        """
        - add or overwrite some parameters of the dictionary
        - create the output directory if does not exist
        """
        pag = False
        ctf = False
        db = None
        if "phase" in _config:
            phase_method = _config["phase"].get("method", "").lower()
            if phase_method in ("pag", "paganin"):
                pag = True
            elif phase_method in ("ctf",):
                ctf = True

            if "delta_beta" in _config["phase"]:
                db = round(float(_config["phase"]["delta_beta"]))
        if "output" in _config:
            _file_name = SingleSliceRunner.get_file_basename_reconstruction(
                scan=self.scan,
                slice_index=self.slice_index,
                pag=pag,
                db=db,
                ctf=ctf,
            )
            _config["output"]["file_prefix"] = "cor_{}_{}".format(_file_name, cor)
            if _config["output"]["location"] not in ("", None):
                # if user specify the location
                if not os.path.isdir(_config["output"]["location"]):
                    os.makedirs(_config["output"]["location"])
            else:
                # otherwise default location will be the data root level
                _config["output"]["location"] = os.sep.join(
                    [
                        self.scan.path,
                        "saaxis_results",
                    ]
                )
        if "reconstruction" not in _config:
            _config["reconstruction"] = {}
        _config["reconstruction"]["start_z"] = self.slice_index
        _config["reconstruction"]["end_z"] = self.slice_index
        return _config

    def preprocess_config(self, config, cor: float):
        dataset_params = self.scan.get_nabu_dataset_info()
        if "dataset" in config:
            dataset_params.update(config["dataset"])
        config["dataset"] = dataset_params

        config["resources"] = utils.get_nabu_resources_desc(
            scan=self.scan, workers=1, method="local"
        )
        # force overwrite results
        if "output" not in config:
            config["output"] = {}
        config["output"].update({"overwrite_results": 1})

        config = self.treateOutputConfig(config, cor=cor)
        # the policy is to save nabu .cfg file at the same location as the
        # force overwrite results

        cfg_folder = os.path.join(
            config["output"]["location"],
            nabu_settings.NABU_CFG_FILE_FOLDER,
        )
        if not os.path.exists(cfg_folder):
            os.makedirs(cfg_folder)

        name = (
            config["output"]["file_prefix"] + nabu_settings.NABU_CONFIG_FILE_EXTENSION
        )
        if not isinstance(self.scan, EDFTomoScan):
            name = "_".join((self.scan.entry.lstrip("/"), name))
        conf_file = os.path.join(cfg_folder, "cor_{}_{}".format(cor, name))
        return config, conf_file



# /*##########################################################################
# Copyright (C) 2016-2017 European Synchrotron Radiation Facility
#
# This file is part of the PyMca X-ray Fluorescence Toolkit developed at
# the ESRF by the Software group.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

__author__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "02/06/2017"

from silx.gui import qt
import numpy.lib.npyio
import os
import tomwer.resources
import logging
from tomwer.io.utils import get_slice_data

logger = logging.getLogger(__name__)


class ImageLoaderThread(qt.QThread):
    """Thread used to load an image"""

    IMG_NOT_FOUND = numpy.load(
        tomwer.resources._resource_filename(
            "%s.%s" % ("imageNotFound", "npy"),
            default_directory=os.path.join("gui", "icons"),
        )
    )

    def __init__(self, url, *args, **kwargs):
        """

        :param index: index of the image on the stackplot
        :param filePath: filePath is the file to load on stackplot reference.
                         It can be an .edf file or a .vol file. If this is a
                         vol file then the name is given with the slice z index
                         to be loaded.
        """
        super().__init__(*args, **kwargs)
        self.data = None
        self.url = url

    def getData(self):
        if hasattr(self, "data"):
            return self.data
        else:
            return None

    def run(self):
        if os.path.exists(self.url.file_path()) and os.path.isfile(
            self.url.file_path()
        ):
            self.data = get_slice_data(self.url)
        else:
            logger.warning("file %s not longer exists or is empty" % self.url)
            self.data = self.IMG_NOT_FOUND



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "04/05/2021"


from ..utils import WidgetLongProcessing
from orangewidget import gui
from orangecontrib.tomwer.orange.managedprocess import SuperviseOW
from orangewidget.settings import Setting
from orangecontrib.tomwer.orange.settings import CallbackSettingsHandler
from orangewidget.widget import Input, Output
from tomwer.core.process.reconstruction.sadeltabeta import SADeltaBetaProcess
from tomwer.synctools.stacks.reconstruction.sadeltabeta import SADeltaBetaProcessStack
from tomwer.synctools.sadeltabeta import QSADeltaBetaParams
from tomwer.synctools.axis import QAxisRP
from silx.gui import qt
from tomwer.gui.reconstruction.sadeltabeta import (
    SADeltaBetaWindow as _SADeltaBetaWindow,
)
from tomwer.core.scan.scanbase import TomwerScanBase, _TomwerBaseDock
from processview.core.manager import ProcessManager, DatasetState
import tomwer.core.process.reconstruction.sadeltabeta.sadeltabeta
from tomwer.core import settings
from tomwer.core import utils
from tomwer.core.cluster import SlurmClusterConfiguration
from tomwer.core.futureobject import FutureTomwerObject
from processview.core import helpers as pv_helpers
import functools
import logging

_logger = logging.getLogger(__name__)


class SADeltaBetaWindow(_SADeltaBetaWindow):
    def __init__(self, Outputs, parent=None, process_id=None):
        _SADeltaBetaWindow.__init__(self, parent=parent)
        self.Outputs = Outputs
        self._sa_delta_beta_params = QSADeltaBetaParams()
        self._processing_stack = SADeltaBetaProcessStack(
            sa_delta_beta_params=self._sa_delta_beta_params, process_id=process_id
        )
        self._clusterConfig = None

    def setClusterConfig(self, cluster_config: dict):
        if not isinstance(
            cluster_config, (dict, type(None), SlurmClusterConfiguration)
        ):
            raise TypeError(
                f"cluster config is expected to be None, dict, {SlurmClusterConfiguration} not {type(cluster_config)}"
            )
        self._clusterConfig = cluster_config

    def _launchReconstructions(self):
        """callback when we want to launch the reconstruction of the
        slice for n cor value"""
        scan = self.getScan()
        if scan is None:
            return
        # step1: if isAutoFocus: validate automatically the scan
        # step2: update the interface if the current scan is the one displayed
        # else skip it
        callback = functools.partial(
            self._mightUpdateResult, scan, self.isAutoFocusLock()
        )
        self._processing_stack.add(
            data=scan, configuration=self.getConfiguration(), callback=callback
        )

    def _validate(self):
        self.validateCurrentScan()

    def _mightUpdateResult(self, scan: TomwerScanBase, validate: bool):
        if not isinstance(scan, TomwerScanBase):
            raise TypeError("scan is expected to be an instance of TomwerScanBase")
        if not isinstance(validate, bool):
            raise TypeError("validate is expected to be a boolean")
        if scan == self.getScan():
            self.setDBScores(
                scan.sa_delta_beta_params.scores,
                score_method=scan.sa_delta_beta_params.score_method,
            )
            if scan.sa_delta_beta_params.autofocus is not None:
                self.setCurrentDeltaBetaValue(scan.sa_delta_beta_params.autofocus)
            ProcessManager().notify_dataset_state(
                dataset=scan,
                process=self._processing_stack,
                details="processing done. Wait for user validation",
                state=DatasetState.WAIT_USER_VALIDATION,
            )
        if validate:
            self.validateScan(scan)

    def wait_processing(self, wait_time):
        self._processing_stack._computationThread.wait(wait_time)

    def validateCurrentScan(self):
        return self.validateScan(self.getScan())

    def validateScan(self, scan):
        if scan is None:
            return
        assert isinstance(scan, TomwerScanBase)
        selected_db_value = (
            self.getCurrentDeltaBetaValue() or scan.sa_delta_beta_params.autofocus
        )
        if selected_db_value is None:
            infos = "no selected delta / beta value. {} skip SADeltaBetaParams".format(
                scan
            )
            _logger.warning(infos)
            scan.sa_delta_beta_params.set_db_selected_value(None)
            pv_helpers.notify_skip(
                process=self._processing_stack, dataset=scan, details=infos
            )
        else:
            scan.sa_delta_beta_params.set_db_selected_value(selected_db_value)
            if scan.nabu_recons_params is not None:
                if "phase" not in scan.nabu_recons_params:
                    scan.nabu_recons_params["phase"] = {}
                scan.nabu_recons_params["phase"]["delta_beta"] = (selected_db_value,)
            infos = "delta / beta selected for {}: {},".format(
                scan, scan.sa_delta_beta_params.value
            )
            pv_helpers.notify_succeed(
                process=self._processing_stack, dataset=scan, details=infos
            )
        SADeltaBetaProcess.process_to_tomwer_processes(
            scan=scan,
        )
        self.Outputs.data.send(scan)

    def getConfiguration(self) -> dict:
        config = super().getConfiguration()
        config["cluster_config"] = self._clusterConfig
        return config

    def setConfiguration(self, config: dict):
        # ignore slurm cluster. Defined by the upper widget
        config.pop("cluster_config", None)
        return super().setConfiguration(config)


class SADeltaBetaOW(SuperviseOW, WidgetLongProcessing):
    """
    Widget for semi-automatic delta / beta calculation

    behavior within a workflow:
    * no delta / beta value will be loaded even if an "axis" window exists on
    the upper stream.
    * if autofocus option is lock:
       * launch the series of reconstruction (with research width defined)
         and the estimated center of rotation if defined. Once the
         reconstruction is ended and if the autofocus button is still lock
         it will select the cor with the highest
         value and mode to workflow downstream.
    * hint: you can define a "multi-step" half-automatic center of rotation
       research by creating several "sa_delta_beta" widget and reducing the research
       width.

    Details about :ref:`sadeltabeta score calculation`
    """

    name = "semi automatic delta/beta calculation"
    id = "orange.widgets.tomwer.sa_delta_beta"
    description = (
        "compute several delta / beta value to get the optimal "
        "delta / beta value (before reconstructing a volume for "
        "example."
    )
    icon = "icons/delta_beta_range.png"
    priority = 22
    keywords = [
        "tomography",
        "semi automatic",
        "half automatic",
        "axis",
        "delta-beta",
        "delta/beta",
        "delta",
        "beta",
        "tomwer",
        "reconstruction",
        "position",
        "center of rotation",
        "sadeltabetaaxis",
        "sa_delta_beta_axis",
        "sa_delta_beta",
        "sadeltabeta",
    ]

    ewokstaskclass = (
        tomwer.core.process.reconstruction.sadeltabeta.sadeltabeta.SADeltaBetaProcess
    )

    want_main_area = True
    resizing_enabled = True
    allows_cycle = True
    compress_signal = False

    settingsHandler = CallbackSettingsHandler()

    sigScanReady = qt.Signal(TomwerScanBase)
    """Signal emitted when a scan is ready"""

    _ewoks_default_inputs = Setting({"data": None, "sa_delta_beta_params": None})

    class Inputs:
        data = Input(name="data", type=TomwerScanBase, default=True, multiple=False)
        data_recompute = Input(
            name="change recons params",
            type=_TomwerBaseDock,
            doc="recompute delta / beta",
        )
        cluster_in = Input(
            name="cluster_config",
            type=SlurmClusterConfiguration,
            doc="slurm cluster to be used",
            multiple=False,
        )

    class Outputs:
        data = Output(name="data", type=TomwerScanBase)

        future_out = Output(
            name="future_tomo_obj",
            type=FutureTomwerObject,
            doc="data with some remote processing",
        )

    def __init__(self, parent=None):
        """

        :param parent: QWidget parent or None
        """
        SuperviseOW.__init__(self, parent)
        WidgetLongProcessing.__init__(self)

        self._layout = gui.vBox(self.mainArea, self.name).layout()
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._widget = SADeltaBetaWindow(
            Outputs=self.Outputs, parent=self, process_id=self.process_id
        )
        self._layout.addWidget(self._widget)

        sa_delta_beta_params = self._ewoks_default_inputs.get(
            "sa_delta_beta_params", None
        )
        self.setConfiguration(sa_delta_beta_params or {})

        # connect signal / slot
        self._widget.sigConfigurationChanged.connect(self._updateSettings)
        self.destroyed.connect(self._widget.stop)
        self._widget._processing_stack.sigComputationStarted.connect(
            self._startProcessing
        )
        self._widget._processing_stack.sigComputationEnded.connect(self._endProcessing)
        # expose API
        self.wait_processing = self._widget.wait_processing

    def __new__(cls, *args, **kwargs):
        # ensure backward compatibility with 'static_input'
        static_input = kwargs.get("stored_settings", {}).get("static_input", None)
        if static_input not in (None, {}):
            _logger.warning(
                "static_input has been deprecated. Will be replaced by _ewoks_default_inputs in the workflow file. Please save the workflow to apply modifications"
            )
            kwargs["stored_settings"]["_ewoks_default_inputs"] = static_input
        return super().__new__(cls, *args, **kwargs)

    def setConfiguration(self, configuration):
        if "workflow" in configuration:
            autofocus_lock = configuration["workflow"].get("autofocus_lock", None)
            if autofocus_lock is not None:
                self._widget.lockAutofocus(autofocus_lock)
            del configuration["workflow"]
        self._widget.setConfiguration(configuration)

    def getCurrentCorValue(self):
        return self._widget.getCurrentCorValue()

    def load_sinogram(self):
        self._widget.loadSinogram()

    def compute(self):
        self._widget.compute()

    def lockAutofocus(self, lock):
        self._widget.lockAutofocus(lock=lock)

    def isAutoFocusLock(self):
        return self._widget.isAutoFocusLock()

    @Inputs.data
    def process(self, scan):
        if scan is None:
            return
        if scan.axis_params is None:
            scan.axis_params = QAxisRP()
        if scan.sa_delta_beta_params is None:
            scan.sa_delta_beta_params = QSADeltaBetaParams()
        self._skipCurrentScan(new_scan=scan)

        if settings.isOnLbsram(scan) and utils.isLowOnMemory(
            settings.get_lbsram_path()
        ):
            self.notify_skip(
                scan=scan,
                details="sa-delta-beta has been skiped for {} because "
                "of low space in lbsram".format(scan),
            )
            self.Outputs.data.send(scan)
        else:
            self._widget.setScan(scan=scan)
            self.notify_pending(scan)
            self.activateWindow()
            if self.isAutoFocusLock():
                self.compute()
            else:
                self.raise_()
                self.show()

    def _skipCurrentScan(self, new_scan):
        scan = self._widget.getScan()
        # if the same scan has been run several scan
        if scan is None or str(scan) == str(new_scan):
            return
        current_scan_state = ProcessManager().get_dataset_state(
            dataset_id=scan.get_identifier(), process=self
        )
        if current_scan_state in (
            DatasetState.PENDING,
            DatasetState.WAIT_USER_VALIDATION,
        ):
            details = "Was pending and has been replaced by another scan."
            self.notify_skip(scan=scan, details=details)
            self.Outputs.data.send(scan)

    @Inputs.data_recompute
    def reprocess(self, dataset):
        self.lockAutofocus(False)
        self.process(dataset)

    @Inputs.cluster_in
    def setCluster(self, cluster):
        self._widget.setClusterConfig(cluster_config=cluster)

    def validateCurrentScan(self):
        self._widget.validateCurrentScan()

    def _updateSettings(self):
        config = self._widget.getConfiguration()
        config.pop("cluster_config", None)
        self._ewoks_default_inputs = {
            "data": None,
            "sa_delta_beta_params": self._widget.getConfiguration(),
        }
        self._ewoks_default_inputs["sa_delta_beta_params"]["workflow"] = {
            "autofocus_lock": self._widget.isAutoFocusLock(),
        }

    def close(self):
        self.stop()
        self._widget = None
        super().close()

    def stop(self):
        self._widget.stop()

    def getConfiguration(self):
        return self._widget.getConfiguration()



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/


__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "02/11/2020"


from typing import Optional
from silx.io.url import DataUrl
from orangewidget.widget import Input, Output
from orangewidget import gui
from tomwer.gui.edit.dkrfpatch import DarkRefPatchWidget
from tomwer.core.scan.hdf5scan import HDF5TomoScan
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.synctools.stacks.edit.darkflatpatch import DarkFlatPatchProcessStack
from orangewidget.settings import Setting
from ...orange.managedprocess import SuperviseOW
from ..utils import WidgetLongProcessing
from tomwer.utils import docstring
import tomwer.core.process.edit.darkflatpatch
import logging

_logger = logging.getLogger(__name__)


class DarkFlatPatchOW(WidgetLongProcessing, SuperviseOW):
    """
    Widget to define on the fly the image_key of a HDF5TomoScan
    """

    name = "dark-flat-patch"
    id = "orange.widgets.tomwer.edit.DarkFlatPatchOW.DarkFlatPatchOW"
    description = "Interface to patch dark and flat to an existing NXTomo" "entry"
    icon = "icons/patch_dark_flat.svg"
    priority = 25
    keywords = [
        "hdf5",
        "nexus",
        "tomwer",
        "file",
        "edition",
        "NXTomo",
        "editor",
        "dark",
        "patch",
        "ref",
        "flat",
    ]

    want_main_area = True
    resizing_enabled = True
    compress_signal = False

    _urlsSetting = Setting(dict())

    ewokstaskclass = tomwer.core.process.edit.darkflatpatch.DarkFlatPatch

    class Inputs:
        data = Input(name="data", type=TomwerScanBase)
        config_in = Input(name="configuration", type=dict)

    class Outputs:
        data = Output(name="data", type=TomwerScanBase)

    def __init__(self, parent=None):
        SuperviseOW.__init__(self, parent=parent)
        WidgetLongProcessing.__init__(self)
        layout = gui.vBox(self.mainArea, self.name).layout()
        self._processingStack = DarkFlatPatchProcessStack(
            self, process_id=self.process_id
        )

        self.widget = DarkRefPatchWidget(parent=self)
        layout.addWidget(self.widget)

        self.setConfiguration(self._urlsSetting)

        # connect signal / slot
        self._processingStack.sigComputationStarted.connect(self._startProcessing)
        self._processingStack.sigComputationEnded.connect(self._endProcessing)
        self.widget.sigConfigurationChanged.connect(self._updateSettings)

    @Inputs.data
    def process(self, scan):
        if scan is None:
            return
        elif not isinstance(scan, HDF5TomoScan):
            _logger.error("We can only patch dark and flat for HDF5TomoScan")
        else:
            self._processingStack.add(scan, self.getConfiguration())

    @docstring(SuperviseOW)
    def reprocess(self, dataset):
        self.process(dataset)

    def getConfiguration(self):
        def cast_url_to_str(url: Optional[DataUrl]):
            if url is None:
                return None
            elif not isinstance(url, DataUrl):
                raise TypeError(
                    f"url is expected to be an optional DataUrl. {type(url)} provided instead"
                )
            else:
                return url.path()

        return {
            "darks_start": cast_url_to_str(self.widget.getStartDarkUrl()),
            "flats_start": cast_url_to_str(self.widget.getStartFlatUrl()),
            "darks_end": cast_url_to_str(self.widget.getEndDarkUrl()),
            "flats_end": cast_url_to_str(self.widget.getEndFlatUrl()),
        }

    @Inputs.config_in
    def setConfiguration(self, config):
        if config is None:
            return
        self.widget.clear()
        url_keys = ("darks_start", "flats_start", "darks_end", "flats_end")
        url_index_keys = (
            "darks_start_index",
            "flats_start_index",
            "darks_end_index",
            "flats_end_index",
        )
        setters = (
            self.widget.setStartDarkUrl,
            self.widget.setStartFlatUrl,
            self.widget.setEndDarkUrl,
            self.widget.setEndFlatUrl,
        )
        for url_key, url_idx_key, setter in zip(url_keys, url_index_keys, setters):
            if url_key in config:
                index = config.get(url_idx_key, 0)
                url = config[url_key]
                if url not in (None, ""):
                    try:
                        setter(url=url, serie_index=index)
                    except Exception as e:
                        _logger.error(e)

    def _endProcessing(self, scan, future_tomo_obj):
        WidgetLongProcessing._endProcessing(self, scan)
        if scan is not None:
            self.Outputs.data.send(scan)

    def _updateSettings(self):
        self._urlsSetting = self.getConfiguration()
        self._urlsSetting.update(
            {
                "darks_start_index": self.widget.getStartDarkIndex(),
                "flats_start_index": self.widget.getStartFlatIndex(),
                "darks_end_index": self.widget.getEndDarkIndex(),
                "flats_end_index": self.widget.getEndFlatIndex(),
            }
        )



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "16/06/2021"


import os
import shutil
import tempfile
import unittest
import pytest
from silx.gui import qt
from tomwer.core import settings
from tomwer.gui.reconstruction.lamino.tofu import TofuWindow
from tomwer.tests.utils import UtilsTest
from tomwer.tests.utils import skip_gui_test


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestLamino(unittest.TestCase):
    """Test the TofuWidget"""

    def setUp(self):
        super().setUp()
        self._window = TofuWindow(parent=None)
        self._widget = self._window._mainWidget
        # add a dark.edf file and a darkHSTFile
        self._folder = tempfile.mkdtemp()

    def tearDown(self):
        super().tearDown()
        self._window.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._window.close()
        shutil.rmtree(self._folder)

    def testSetup(self):
        """Make sure the widget is correctly setting parameters from a scan"""
        dataset = UtilsTest.getEDFDataset("test01")
        self._window.loadFromScan(dataset)
        reconsParams = self._widget.getParameters()
        # check general information
        # note : the .info file is incorrect
        self.assertTrue(self._widget._tabs._inputWidget.getNumber() == 1)
        self.assertTrue(reconsParams["number"] == 1)
        # check flat field correction is set
        ffcWidget = self._widget._tabs._inputWidget.widget._ffcWidget
        thFlats = os.path.join(dataset, "refHST0000.edf")
        self.assertTrue(ffcWidget.getFlats() == thFlats)
        self.assertTrue(reconsParams["flats"] == thFlats)
        thSecondFlats = os.path.join(dataset, "refHST0010.edf")
        self.assertTrue(ffcWidget.getSecondFlats() == thSecondFlats)
        self.assertTrue(reconsParams["flats2"] == thSecondFlats)
        thDarks = os.path.join(dataset, "darkend0000.edf")
        self.assertTrue(ffcWidget.getDarks() == thDarks)
        self.assertTrue(reconsParams["darks"] == thDarks)

        inputWidget = self._widget._tabs._inputWidget.widget
        self.assertTrue(inputWidget.rotationAngle._overallAngle.getAngle() == 180.0)
        self.assertTrue(inputWidget.rotationAngle._axisAngleY.getAngle() == 0.0)
        self.assertTrue(inputWidget.rotationAngle._axisAngleZ.getAngle() == 0.0)

    def testSetupWithLockedParams(self):
        """Make sure lockers are corrcetly taking into account"""
        # check flat field correction is set

        inputWidget = self._widget._tabs._inputWidget.widget
        inputWidget.rotationAngle._overallAngle.setAngle(12.0)
        inputWidget.rotationAngle._overallAngle.lock()

        inputWidget.rotationAngle._axisAngleY.setAngle(-26.0)
        inputWidget.rotationAngle._axisAngleY.lock()

        inputWidget.rotationAngle._axisAngleZ.setAngle(63.5)
        inputWidget.rotationAngle._axisAngleZ.lock()

        outputWidget = self._widget._tabs._outputWidget
        outputWidget.setOutput("toto")
        outputWidget.lockOutput()

        dataset = UtilsTest.getEDFDataset("test01")
        self._window.loadFromScan(dataset)

        self.assertTrue(inputWidget.rotationAngle._overallAngle.getAngle() == 12.0)
        self.assertTrue(inputWidget.rotationAngle._axisAngleY.getAngle() == -26.0)
        self.assertTrue(inputWidget.rotationAngle._axisAngleZ.getAngle() == 63.5)

        self.assertTrue(outputWidget.getOutput() == "toto")

    def testEditionDisplay(self):
        """
        Test that edition of parameters by the user are taking into account
        """
        displayWidget = self._widget._tabs._inputWidget.widget.rotationAngle
        # angles
        displayWidget._laminoAngle.setAngle(12.0)
        self.assertTrue(self._widget.getParameters()["axis-angle-x"] == 12.0)
        displayWidget._axisAngleY.setAngle(10.0)
        self.assertTrue(self._widget.getParameters()["axis-angle-y"] == 10.0)

        # regions
        regionWidget = self._widget._tabs._outputWidget._region
        regionWidget._xRegion._regionLE.setText("0")
        self.assertTrue(self._widget.getParameters()["x-region"] == (0, -1, 1))
        regionWidget.setYRegion(10, 12, 25)
        self.assertTrue(self._widget.getParameters()["y-region"] == (10, 12, 25))

        # make sure saving and loading is correct
        savingFile = os.path.join(self._folder, "savingFile.ini")
        self._widget.saveTo(savingFile)
        self.assertTrue(os.path.exists(savingFile))
        regionWidget.setYRegion(10, 12, 22)
        self.assertTrue(self._widget.getParameters()["y-region"] == (10, 12, 22))
        self._widget.loadFile(savingFile)
        self.assertTrue(self._widget.getParameters()["y-region"] == (10, 12, 25))

    def testOutputWidget(self):
        """test behavior of the output widget and link with get and set
        parameters"""
        outputWidget = self._widget._tabs._outputWidget
        # volume angles
        outputWidget._volumeAngleGrp.setAngles(0, 20, -36.0)
        reconsParams = self._widget.getParameters()
        self.assertTrue(reconsParams["volume-angle-x"] == 0)
        self.assertTrue(reconsParams["volume-angle-y"] == 20.0)
        self.assertTrue(reconsParams["volume-angle-z"] == -36.0)

    def testCenteringWidget(self):
        """test behavior of the centering widget and link with get and set
        parameters"""
        centeringWidget = self._widget._tabs._inputWidget.widget.centeringWidget
        centeringWidget._xCenterLE.setText("12")
        centeringWidget._yCenterLE.setText("2")
        self.assertTrue(self._widget.getParameters()["center-position-x"] == 12.0)
        self.assertTrue(self._widget.getParameters()["center-position-z"] == 2.0)

    def testForceLbsRam(self):
        """Test the behavior of the forcelbsram option"""
        scan1 = tempfile.mkdtemp(suffix="1")
        self._window.loadFromScan(scan1)
        _end = os.path.join("1", "xySlice")
        self.assertTrue(self._window.getParameters()["output"].endswith(_end))

        self._window._mainWidget._tabs._outputWidget.lockOutput()
        scan2 = tempfile.mkdtemp(suffix="2")
        self._window.loadFromScan(scan2)
        _end = os.path.join("1", "xySlice")
        self.assertTrue(self._window.getParameters()["output"].endswith(_end))

        self._window._mainWidget._tabs._outputWidget.lockOutput(False)
        self._window.loadFromScan(scan2)
        _end = os.path.join("2", "xySlice")
        self.assertTrue(self._window.getParameters()["output"].endswith(_end))

        self._window._mainWidget._tabs._outputWidget.setForceLbsram(True)
        self.assertTrue(
            self._window.getParameters()["output"].startswith(
                settings.get_lbsram_path()
            )
        )
        self._window.loadFromScan(scan1)
        self.assertTrue(
            self._window.getParameters()["output"].startswith(
                settings.get_lbsram_path()
            )
        )



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "01/06/2018"


import glob
import os
import subprocess
import sys
import enum
from collections import OrderedDict
import tempfile
import shutil
from tomwer.core.scan.hdf5scan import HDF5TomoScan
from tomwer.core.process.task import Task
from tomwer.core.process.reconstruction.darkref.settings import (
    DARKHST_PREFIX,
    REFHST_PREFIX,
)
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.utils import getDim1Dim2
from tomwer.core.utils.char import PSI_CHAR
import tomwer.version

import logging

logger = logging.getLogger(__name__)


@enum.unique
class FFCWhen(enum.Enum):
    on_the_fly = 0
    preprocessing = 1


SLICE_STACK_TYPE = "slice stack"
ROTATION_CENTER_TYPE = "rotation center"
LAMINO_ANGLE_TYPE = "lamino angle"
PSI_ANGLE_TYPE = " ".join((PSI_CHAR, "angle"))

SCAN_TYPES = OrderedDict(
    (
        (SLICE_STACK_TYPE, "z"),
        (ROTATION_CENTER_TYPE, "center-position-x"),
        (LAMINO_ANGLE_TYPE, "axis-angle-x"),
        (PSI_ANGLE_TYPE, "axis-angle-y"),
    )
)

SCAN_TYPES_I = {}
for s_type, value in SCAN_TYPES.items():
    SCAN_TYPES_I[value] = s_type


def _retrieve_opts_recons_cmd(scan_id, recons_param, additional_opts, pre_proc_ffc):
    """Return the options command under the style '--name optValue' from the
    ToFuReconstructionParam and a string of additional options

    :param bool pre_proc_ffc: should we apply flat field correction on the fly
                              or is it already done.
    """
    assert isinstance(recons_param, dict)
    options = [""]
    # deal with option that will can be None
    for opt in (
        "retry-timeout",
        "retries",
        "reduction-mode",
        "dark-scale",
        "darks",
        "slices-per-device",
        "slice-memory-coeff",
        "number",
        "volume-angle-x",
        "volume-angle-y",
        "volume-angle-z",
        "output",
        "center-position-x",
        "center-position-z",
        "axis-angle-x",
        "axis-angle-y",
        "z-parameter",
        "retrieval-method",
        "energy",
        "pixel-size",
        "propagation-distance",
        "regularization-rate",
        "thresholding-rate",
        "flats",
        "flats2",
        "z",
    ):
        if recons_param[opt] is not None:
            # special case for option that can contain Unix filename pattern matching (*)
            if opt in ("darks", "flats", "flats2", "dark-scale"):
                if pre_proc_ffc is True:
                    # in this case flat field correction has already been run
                    pass
                else:
                    options = options + [
                        " ".join([opt, '"' + str(recons_param[opt]) + '"'])
                    ]
            else:
                options = options + [" ".join([opt, str(recons_param[opt])])]

    # deal with specific case of the scan
    if "projections" in recons_param:
        options = options + [recons_param["projections"]]
    elif scan_id is not None:
        concert_radio_path = os.path.join(scan_id, "radios")
        proj_file_fn = concert_radio_path + os.sep + "frame_*.tif"
        if os.path.exists(concert_radio_path) and len(glob.glob(proj_file_fn)) > 0:
            options = options + ['projections "' + proj_file_fn + '"']
        else:
            options = options + [
                'projections "'
                + scan_id
                + os.sep
                + os.path.basename(scan_id)
                + '*.edf"'
            ]

    # deal with region
    assert "region" in recons_param
    if recons_param["region"] is not None:
        assert type(recons_param["region"]) in (list, tuple)
        assert len(recons_param["region"]) == 3
        options = options + ["region=%s,%s,%s" % recons_param["region"]]

    # deal with overallangle wich need a '-' for avoid staring by a numerical
    assert type(recons_param["overall-angle"]) is float
    options = options + ["overall-angle " + str(-1.0 * recons_param["overall-angle"])]

    # deal with x and y region
    for region in ("x-region", "y-region"):
        value = recons_param[region]
        if value is not None:
            value = str(value).replace(" ", "")
            value = value.lstrip("(").rstrip(")")
            options = options + ["=".join((region, value))]

    # deal with option that will be defined only if set to true
    for opt in ("verbose", "dry-run", "absorptivity"):
        assert opt in recons_param
        if recons_param[opt] is True:
            options.append(opt)

    return " --".join(options) + " " + additional_opts


# TODO: additional opts should be specific to flat field correction.
# reconstruction and ffc additional options should be separated
def _retrieve_opts_ffc_cmd(scan, recons_param, additional_opts, output):
    """Return the options command under the style '--name optValue' from the
    ToFuReconstructionParam and a string of additional options

    :param str output: output directory
    """
    assert isinstance(scan, TomwerScanBase)
    options = [""]

    options = options + [" ".join(["output", output])]

    # deal with option that will can be None
    for opt in (
        "retry-timeout",
        "retries",
        "reduction-mode",
        "dark-scale",
        "darks",
        "number",
        "center-position-x",
        "center-position-z",
        "axis-angle-x",
        "retrieval-method",
        "energy",
        "pixel-size",
        "propagation-distance",
        "regularization-rate",
        "thresholding-rate",
        "flats",
        "flats2",
    ):
        if recons_param[opt] is not None:
            # special case for option that can contain Unix filename pattern matching (*)
            if opt in ("darks", "flats", "flats2"):
                options = options + [
                    " ".join([opt, '"' + str(recons_param[opt] + '"')])
                ]
            else:
                options = options + [" ".join([opt, str(recons_param[opt])])]

    # deal with specific case of the scanID
    if scan.path is not None:
        options = options + [_get_projection_file_pattern(scan_path=scan.path)]

    # deal with option that will be defined only if set to true
    for opt in ("verbose", "dry-run", "absorptivity"):
        assert opt in recons_param
        # absorptivity should only be done in reconstruction
        if opt == "absorptivity":
            continue
        if recons_param[opt] is True:
            options.append(opt)

    return " --".join(options) + " " + additional_opts


def _get_projection_file_pattern(scan_path, ffc_folder=None):
    """

    :param scan_path: original scan path
    :param ffc_folder: folder containing the preprocessed flat field corrected
                       image
    :return: projection + name of the scan path
    """
    assert type(scan_path) is str
    if ffc_folder is None:
        root_folder = scan_path
    else:
        root_folder = ffc_folder
    assert root_folder is not None
    concert_radio_path = os.path.join(root_folder, "radios")
    proj_file_fn = concert_radio_path + os.sep + "frame_*.tif"
    if os.path.exists(concert_radio_path) and len(glob.glob(proj_file_fn)) > 0:
        return 'projections "' + proj_file_fn + '"'
    else:
        return (
            'projections "'
            + root_folder
            + os.sep
            + os.path.basename(scan_path)
            + '*.edf"'
        )


def _tofu_lamino_reconstruction(
    scan_id,
    recons_param,
    additional_options,
    delete_existing,
    exec_cmd=True,
    pre_proc_ffc=True,
):
    """Process a reconstruction for lamino using tofu

    :param str scan_id: path of the scan to reconstruct
    :param dict recons_param: parameters for the reconstruction
    :param str additional_options: additional options to be add at the tofu reco
                                  call.
    :param bool delete_existing: if True then remove output dir if given
    :param bool exec_cmd: if True, will run reconstruction, otherwise will only
                          display the reconstruction parameters.
    :param bool pre_proc_ffc: should we apply flat field correction on the fly
                              or is it already done. If flat field correction
                              has been preprocessed, also deal with half
                              acquisition case
    """
    assert "output" in recons_param
    outputdir = recons_param["output"]

    if exec_cmd is True:
        if has_tofu() is False:
            logger.error(
                "Cannot launch tofu reconstruction because " "tofu is not installed."
            )
            return
    if delete_existing is True and "output" in recons_param:
        logger.info("remove output dir: %s" % recons_param["output"])
        if exec_cmd:
            for _file in glob.glob(outputdir + "*.tif"):
                try:
                    os.remove(_file)
                except Exception as e:
                    logger.error(e)
        else:
            logger.info(("will remove all files" + outputdir + "*.tif"))

    options = _retrieve_opts_recons_cmd(
        scan_id=scan_id,
        recons_param=recons_param,
        additional_opts=additional_options,
        pre_proc_ffc=pre_proc_ffc,
    )

    try:
        logger.info("launch command : " + "tofu reco" + options)
        if exec_cmd is True:
            subprocess.call(
                "tofu reco " + options, shell=True, stderr=sys.stderr, stdout=sys.stdout
            )
    except OSError:
        return False
    else:
        return True


def _preprocess_ffc(scan, recons_param, additional_options, output, exec_cmd=True):
    """Process a flat field reconstruction for given scan using tofu

    :param TomwerScanBase scan: path of the scan to reconstruct
    :param dict recons_param: parameters for the reconstruction
    :param str additional_options: additional options to be add at the tofu reco
                                  call.
    :param bool exec_cmd: if True, will run reconstruction, otherwise will only
                          display the reconstruction parameters.
    """
    assert isinstance(scan, TomwerScanBase)
    if exec_cmd is True:
        if has_tofu() is False:
            logger.error(
                "Cannot launch tofu reconstruction because " "tofu is not installed."
            )
            return

    output_folder = os.path.join(output, "fc")
    if exec_cmd is True and os.path.exists(output_folder):
        logger.info("removing" + output_folder)
        shutil.rmtree(output_folder)
    output = output_folder + "-%04i.tif"
    options = _retrieve_opts_ffc_cmd(
        scan=scan,
        recons_param=recons_param,
        additional_opts=additional_options,
        output=output,
    )
    if output is not None:
        logger.info(
            "flat field correction preprocessing result will be store in " + output
        )
    logger.info("launch command : " + "tofu preprocess" + options)
    if exec_cmd is True:
        subprocess.call(
            "tofu preprocess " + options,
            shell=True,
            stderr=sys.stderr,
            stdout=sys.stdout,
        )
    return True


def _preprocess_stitching(
    shift, blend, adjust_mean, fc_folder, fc1_folder, fc2_folder, dry_run
):
    """
    move half acquisition dataset to full within flips

    :param float shift: How much is second image shifted with respect to the
                        first one. For example, shift 0 means that both images
                        overlap perfectly and the stitching doesn’t actually
                        broaden the image. Shift corresponding to image width
                        makes for a stitched image with twice the width of the
                        respective images.
    :param bool blend: Linearly interpolate between the two images in the
                       overlapping region.
    :param bool adjust_mean: Compute the mean of the overlapping region in the
                             two images and adjust the second image to match
                             the mean of the first one.
    :param str fc_folder: original folder of the sinogram
    :param str fc1_folder:
    :param str fc2_folder:
    :param bool dry_run:
    """
    assert shift is not None
    assert blend is not None
    if dry_run is False:
        if os.path.exists(fc1_folder):
            logger.info("removing" + fc1_folder)
            shutil.rmtree(fc1_folder)
        os.makedirs(fc1_folder)
        if os.path.exists(fc2_folder):
            logger.info("removing" + fc2_folder)
            shutil.rmtree(fc2_folder)
        os.makedirs(fc2_folder)

    def move_tiff():
        n_file = len(os.listdir(fc_folder))
        if n_file % 2 != 0:
            logger.error("even number of projection file, unable to apply stitching")
            return False
        order_list_dir = os.listdir(fc_folder)
        order_list_dir.sort()
        for i_file, file_ in enumerate(list(order_list_dir)):
            if i_file < n_file / 2:
                dest = fc1_folder
            else:
                dest = fc2_folder
            file_path = os.path.join(fc_folder, file_)
            shutil.move(src=file_path, dst=dest)
        # now fc should be empty of any tif file
        return True

    def stitching(shift, blend, adjust_mean):
        work = "[read path=%s, read path=%s ! flip direction=horizontal]" % (
            fc1_folder,
            fc2_folder,
        )
        work = work + " ! stitch shift=%s " % shift
        work += "blend=%s " % str(int(blend))
        work += "adjust-mean=%s " % str(int(adjust_mean))
        work += "! write filename=%s" % os.path.join(fc_folder, "fc-%04i.tif")
        try:
            logger.info("launch command : " + "ufo-launch " + work)
            if not dry_run:
                subprocess.call(
                    "ufo-launch " + work,
                    shell=True,
                    stderr=sys.stderr,
                    stdout=sys.stdout,
                )
        except OSError:
            return False
        else:
            return True

    if not dry_run:
        move_res = move_tiff()
    else:
        logger.info("move tiff file from fc to fc1, fc2")
        move_res = True
    if move_res is True:
        stitching(shift, blend=blend, adjust_mean=adjust_mean)
    return fc_folder


def has_tofu():
    """

    :return: true if the os knows the tofu command
    """
    if not sys.platform.startswith("linux"):
        return False
    try:
        subprocess.call(
            ["tofu", "--version"], stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        return True
    except Exception:
        return False


class LaminoReconstruction(Task, input_names=("data",), output_names=("data",)):
    """
    Process to launch the lamino reconstruction

    TODO: setting parameters should be group with the gui/lamino/tofu/xxx files
    But this binding is probably overkilled as we should use the tofu python
    binding for it. But no python3 version at the moment.
    """

    DEFAULT_RECONSTRUCTION_PARAMETERS_VALS = {
        "region": None,
        "retry-timeout": None,
        "retries": None,
        "reduction-mode": None,
        "dark-scale": None,
        "darks": None,
        "slices-per-device": None,
        "slice-memory-coeff": 0.8,
        "number": None,
        "volume-angle-x": None,
        "volume-angle-y": None,
        "volume-angle-z": None,
        "output": None,
        "center-position-x": None,
        "center-position-z": None,
        "axis-angle-x": None,
        "axis-angle-y": None,
        "z-parameter": None,
        "retrieval-method": None,
        "energy": None,
        "pixel-size": None,
        "propagation-distance": None,
        "regularization-rate": None,
        "thresholding-rate": None,
        "flats": None,
        "flats2": None,
        "overall-angle": 360.0,
        "x-region": None,
        "y-region": None,
        "verbose": False,
        "dry-run": False,
        "absorptivity": False,
        "half-acquisition": False,
        "z": None,
    }

    def __init__(
        self, varinfo=None, inputs=None, node_id=None, node_attrs=None, execinfo=None
    ):
        Task.__init__(
            self,
            varinfo=varinfo,
            inputs=inputs,
            node_id=node_id,
            node_attrs=node_attrs,
            execinfo=execinfo,
        )
        if "recons_params" in inputs:
            raise KeyError(
                "Do not use recons_params but `lamino_recons_params` instead"
            )

        self._reconsparams = self.DEFAULT_RECONSTRUCTION_PARAMETERS_VALS
        recons_params = inputs.get("lamino_params", {})
        self._reconsparams.update(recons_params)
        self._additional_reco_options = ""
        self._additional_preprocess_options = ""
        self._delete_existing = False
        self._dry_run = False
        self.__ffc_has_been_preprocessed = False
        self.__ffc_tmp_dir = None
        self.__prepare_ffc_dir()
        # top level directory to make the stitching and flat field preprocessing

        self.__latest_ffc_prerecons = (None, None, None, None, None, self._dry_run)
        # Store the current existing flat field reconstruction as the 'ffc key'
        # (scan, x center, method, dark, ff, dry_run) because this should be
        # executed only if x center change or scan path and we want to keep the same
        # behavior with or without the dry-run option
        self.__latest_stitching = None, None

    def __del__(self):
        self.__remove_ffc_dir()

    def __remove_ffc_dir(self):
        if self.__ffc_tmp_dir is not None and os.path.exists(self.__ffc_tmp_dir):
            shutil.rmtree(self.__ffc_tmp_dir)

    def __prepare_ffc_dir(self):
        self.__remove_ffc_dir()
        self.__ffc_tmp_dir = tempfile.mkdtemp()
        os.makedirs(os.path.join(self.__ffc_tmp_dir, "fc"))

    @property
    def reconstruction_parameters(self):
        return self._reconsparams

    @reconstruction_parameters.setter
    def reconstruction_parameters(self, params):
        self._reconsparams = params
        if "output" in self._reconsparams:
            self.dry_run = self._reconsparams["output"] in (None, "")

    @property
    def additional_reco_options(self):
        return self._additional_reco_options

    @additional_reco_options.setter
    def additional_reco_options(self, opts):
        self._additional_reco_options = opts
        if "delete-existing" in self.additional_reco_options:
            self.delete_existing = self.additional_reco_options["delete-existing"]

    @property
    def additional_preprocess_options(self):
        return self._additional_preprocess_options

    @additional_preprocess_options.setter
    def additional_preprocess_options(self, opts):
        self._additional_preprocess_options = opts

    @property
    def delete_existing(self):
        return self._delete_existing

    @delete_existing.setter
    def delete_existing(self, _delete):
        assert type(_delete) is bool
        self._delete_existing = _delete

    @property
    def dry_run(self):
        return self._dry_run

    @dry_run.setter
    def dry_run(self, dryrun):
        assert type(dryrun) is bool
        self._dry_run = dryrun

    def preprocess_ff(self, scan):
        """preprocess flat field corrcetion on the given scan.
        Result will be stored in `__ffc_tmp_dir` and the tuple
        (scan, center-position-x) will be stored as a key to avoid repeating
        this preprocessing flat field if already process.

        :return: path of the flat field corrected file or None if cannot process
        """
        assert isinstance(scan, TomwerScanBase)
        if scan is None:
            return None

        def get_val_or_none(key):
            if key in self._reconsparams:
                return self._reconsparams[key]
            else:
                return None

        x_center = get_val_or_none("center-position-x")

        if "output" not in self._reconsparams:
            logger.error(
                "no output define, requested for pre processing the "
                "flat field correction in pre processing."
            )
            return None
        elif x_center is None:
            logger.error(
                "x center position not defined, unable to process flat "
                "field correction"
            )
            return None

        output_folder = self._reconsparams["output"]
        output_folder = self._get_fc_folder_frm_output_folder(output_folder)
        fc_folder = os.path.join(output_folder, "fc")
        # TODO: should not integrate xySlice ...

        # case data is already stored in ffc_preprocess_dir
        darks = get_val_or_none("darks")
        flats = get_val_or_none("flat")
        flats2 = get_val_or_none("flats2")
        method = get_val_or_none("reduction-mode")
        if (
            scan,
            x_center,
            method,
            darks,
            [flats, flats2],
            self._dry_run,
        ) == self.__latest_ffc_prerecons:
            logger.info("flat field correction already process, skip correction")
            return fc_folder

        recons_params = self.reconstruction_parameters
        recons_params["projections"] = _get_projection_file_pattern(scan_path=scan.path)
        _preprocess_ffc(
            scan=scan,
            recons_param=recons_params,
            additional_options=self.additional_preprocess_options,
            exec_cmd=(not self.dry_run),
            output=fc_folder,
        )
        recons_params["projections"] = os.path.join(fc_folder, "*.tif")
        self.__latest_ffc_prerecons = (
            scan,
            x_center,
            method,
            darks,
            [flats, flats2],
            self._dry_run,
        )
        return fc_folder

    def _get_fc_folder_frm_output_folder(self, output_folder):
        if output_folder is None:
            return "/ghost_folder"
        else:
            return os.path.dirname(output_folder)

    def is_ffc_has_been_preprocessed(self, scan, x_center, method, darks, ff):
        return (
            scan,
            x_center,
            method,
            darks,
            ff,
            self._dry_run,
        ) == self.__latest_ffc_prerecons

    def stitching_requested(self):
        """

        :return: True if stitching is requested
        """
        return (
            "half-acquisition" in self._reconsparams
            and self._reconsparams["half-acquisition"] is True
        )

    def need_reprocessing_stitching(
        self, shift, blend, adjust_mean, fc_folder, fc1_folder, fc2_folder, dry_run
    ):
        """

        :param float shift:
        :param bool blend:
        :param bool adjust_mean:
        :param str fc_folder:
        :param str fc1_folder:
        :param str fc2_folder:
        :param bool dry_run:
        :return: True if stitching need to be reprocess
        """
        return self.__latest_stitching != (
            self.__latest_ffc_prerecons,
            (shift, blend, adjust_mean, fc_folder, fc1_folder, fc2_folder, dry_run),
        )

    def preprocessing_requested(self):
        """

        :return: True if some preprocessing is require
        """
        return (
            "ffc-when" in self._reconsparams
            and self._reconsparams["ffc-when"] is FFCWhen.preprocessing
        )

    def need_reprocessing_ffc(self, scan, x_center, method, darks, ff, dry_run):
        """

        :param :class:`.TomoBase` scan: scan to process
        :param float x_center: x center
        :param str method: can be 'median' or 'average'
        :param str darks:
        :param list ff:
        :param bool dry_run:
        :return: True if flat field need to be reprocess
        """
        assert isinstance(scan, TomwerScanBase)
        return self.__latest_ffc_prerecons != (
            scan.path,
            x_center,
            method,
            darks,
            ff,
            dry_run,
        )

    def run(self):
        scan = self.inputs.data
        if scan is None:
            self.outputs.data = None
            return
        if type(scan) is dict:
            _scan = ScanFactory.create_scan_object_frm_dict(scan)
        elif type(scan) is str:
            _scan = ScanFactory.create_scan_object(scan_path=scan)
        else:
            _scan = scan
        if not isinstance(_scan, TomwerScanBase):
            raise TypeError(
                "input data is expected to be dict, str or "
                "TomwerScanBase not {}".format(type(_scan))
            )

        # if need some preprocessing
        if self.preprocessing_requested():
            print("------------------------")
            print("request preprocessing ffc")
            self._ff_has_been_reprocess = False
            # if need to reprocess flat field
            ffc_key = {
                "scan": _scan,
                "x_center": self._reconsparams["center-position-x"],
                "method": self._reconsparams["reduction-mode"],
                "darks": self._reconsparams["darks"],
                "ff": [self._reconsparams["flats"], self._reconsparams["flats2"]],
                "dry_run": self._dry_run,
            }
            if self.need_reprocessing_ffc(**ffc_key):
                ffc_folder = self.preprocess_ff(scan=_scan)
                if ffc_folder is None:
                    logger.error("Fail to process the flat field correction")
                    return None
                self._reconsparams["projections"] = _get_projection_file_pattern(
                    scan_path=scan.path, ffc_folder=ffc_folder
                )
                self._ff_has_been_reprocess = True

            if self.stitching_requested():
                output_folder = self._reconsparams["output"]
                output_folder = self._get_fc_folder_frm_output_folder(output_folder)
                fc_folder = os.path.join(output_folder, "fc")
                fc1_folder = os.path.join(output_folder, "fc1")
                fc2_folder = os.path.join(output_folder, "fc2")
                shift = self._reconsparams["center-position-x"]
                blend = self._reconsparams["blend"]
                adjust_mean = self._reconsparams["adjust-mean"]
                if self.need_reprocessing_stitching(
                    shift=shift,
                    fc_folder=fc_folder,
                    fc1_folder=fc1_folder,
                    fc2_folder=fc2_folder,
                    dry_run=self.dry_run,
                    blend=blend,
                    adjust_mean=adjust_mean,
                ):
                    print("------------------------")
                    print("run stitching")
                    image_width = getDim1Dim2(scan.path)[0]
                    if image_width is None:
                        logger.warning(
                            "failed to find image width, set to " "2048 by default"
                        )
                        image_width = 2048
                    shift = 2.0 * self._reconsparams["center-position-x"] - image_width
                    _preprocess_stitching(
                        shift=shift,
                        fc_folder=fc_folder,
                        fc1_folder=fc1_folder,
                        fc2_folder=fc2_folder,
                        dry_run=self.dry_run,
                        blend=blend,
                        adjust_mean=adjust_mean,
                    )

                    self.__latest_stitching = (
                        self.__latest_ffc_prerecons,
                        (
                            shift,
                            blend,
                            adjust_mean,
                            fc_folder,
                            fc1_folder,
                            fc2_folder,
                            self.dry_run,
                        ),
                    )
                else:
                    logger.info("data already stitch from previous process")
                fc_folder = os.path.join(output_folder, "fc")
                # stitching reduce the number of projections by a factor of 2
                self._reconsparams["number"] = int(self._reconsparams["number"] / 2)
                self._reconsparams["overall-angle"] = (
                    self._reconsparams["overall-angle"] / 2.0
                )
            else:
                # TODO: remove, ffc folder is fc folder
                fc_folder = ffc_folder
            self._reconsparams["projections"] = 'projections "' + os.path.join(
                fc_folder, '*.tif"'
            )

        res = _tofu_lamino_reconstruction(
            scan_id=_scan.path,
            recons_param=self._reconsparams,
            additional_options=self.additional_reco_options,
            delete_existing=self.delete_existing,
            exec_cmd=(not self.dry_run),
            pre_proc_ffc=self.preprocessing_requested(),
        )
        if not self.dry_run:
            entry = "entry"
            if isinstance(scan, HDF5TomoScan):
                entry = scan.entry
            with scan.acquire_process_file_lock():
                self.register_process(
                    process_file=scan.process_file,
                    entry=entry,
                    configuration=self._reconsparams,
                    results={},
                    process_index=scan.pop_process_index(),
                    overwrite=True,
                )

        if res is False:
            logger.error(f"Reconstruction of {_scan.path} failed")
        if self._return_dict:
            self.outputs.data = _scan.to_dict()
        else:
            self.outputs.data = _scan

    def set_configuration(self, configuration):
        assert isinstance(configuration, (tuple, list))
        self.additional_preprocess_options, self.additional_reco_options = configuration

    @staticmethod
    def program_name():
        return "lamino"

    @staticmethod
    def program_version():
        """version of the program used for this processing"""
        return tomwer.version.version


def getDark(scan):
    """Return darks as a string for tofu from the scan path"""
    concert_projection_files = os.path.join(scan, "radios")
    concert_darks_path = os.path.join(scan, "darks")
    if (
        os.path.exists(concert_darks_path)
        and os.path.exists(concert_projection_files)
        and len(glob.glob(concert_darks_path + os.sep + "frame_*.tif")) > 0
    ):
        return concert_darks_path + os.sep + "frame_*.tif"
    else:
        files = os.listdir(scan)
        for thFile in (
            DARKHST_PREFIX,
            "dark.edf",
            "darkend0000.edf",
            "darkend000.edf",
            "darkHST.edf",
        ):
            if thFile in files:
                return os.path.join(scan, thFile)
        return None


def getFlats(scan):
    """
    Return flats as a string for tofu from the scan path

    :return: tuple (flats, secondFlats)
    """

    def treatRawRef(rawRefFiles):
        ns = set()
        for _file in rawRefFiles:
            name = _file.rstrip(".edf")
            ns.add(name.split("_")[-1])
        ns = sorted(ns)
        if len(ns) == 0:
            return None, None
        elif len(ns) == 1:
            return os.path.join(scan, "ref*_" + ns[0] + ".edf")
        else:
            return (
                os.path.join(scan, "ref*_" + ns[0] + ".edf"),
                os.path.join(scan, "ref*_" + ns[-1] + ".edf"),
            )

    # deal with concert files
    concert_projection_files = os.path.join(scan, "radios")
    concert_flats_path = os.path.join(scan, "flats")
    if os.path.exists(concert_flats_path) and os.path.exists(concert_projection_files):
        flats = None
        flats2 = None
        if len(glob.glob(concert_flats_path + os.sep + "frame_*.tif")) > 0:
            flats = concert_flats_path + os.sep + "frame_*.tif"
        concert_flats2_path = os.path.join(scan, "flats_2")
        if len(glob.glob(concert_flats2_path + os.sep + "frame_*.tif")) > 0:
            flats2 = concert_flats2_path + os.sep + "frame_*.tif"
        if flats is not None or flats2 is not None:
            return flats, flats2

    # deal with classical files
    files = os.listdir(scan)
    refHSTFiles = []  # files starting with refHST (treated by darkRef)
    rawRefFiles = []  # files starting with ref only (raw ref files)
    for _file in files:
        if _file.startswith("ref") and _file.endswith(".edf"):
            if _file.startswith(REFHST_PREFIX):
                refHSTFiles.append(_file)
            else:
                rawRefFiles.append(_file)

    refHSTFiles = sorted(refHSTFiles)
    if len(refHSTFiles) == 0:
        return treatRawRef(rawRefFiles)
    elif len(refHSTFiles) == 1:
        return os.path.join(scan, refHSTFiles[0]), None
    else:
        logger.warning("Found more than two refHST files")
        return (os.path.join(scan, refHSTFiles[0]), os.path.join(scan, refHSTFiles[-1]))



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "14/12/2021"


from orangewidget.workflow import config
import tomwer.version
from .splash import splash_screen, getIcon
from tomwer.gui import icons
import pkg_resources
from silx.gui import qt
from orangecanvas.application.outputview import (
    TerminalTextDocument as _TerminalTextDocument,
)
from . import environ
from orangewidget.settings import set_widget_settings_dir_components
from .widgetsscheme import WidgetsScheme


def version():
    return tomwer.version.version


class TomwerConfig(config.Config):
    """
    Configuration defined for tomwer
    """

    OrganizationDomain = "esrf"
    ApplicationName = "tomwer"
    ApplicationVersion = version()

    def init(self):
        super().init()
        qt.QApplication.setApplicationName(self.ApplicationName)
        qt.QApplication.setOrganizationDomain(self.OrganizationDomain)
        qt.QApplication.setApplicationVersion(self.ApplicationVersion)
        qt.QApplication.setApplicationDisplayName(self.ApplicationName)
        widget_settings_dir_cfg = environ.get_path("widget_settings_dir", "")
        if widget_settings_dir_cfg:
            # widget_settings_dir is configured via config file
            set_widget_settings_dir_components(
                widget_settings_dir_cfg, self.ApplicationVersion
            )

        canvas_settings_dir_cfg = environ.get_path("canvas_settings_dir", "")
        if canvas_settings_dir_cfg:
            # canvas_settings_dir is configured via config file
            qt.QSettings.setPath(
                qt.QSettings.IniFormat, qt.QSettings.UserScope, canvas_settings_dir_cfg
            )

    @staticmethod
    def splash_screen():
        return splash_screen()

    @staticmethod
    def core_packages():
        return super().core_packages() + ["tomwer-add-on"]

    @staticmethod
    def application_icon():
        return getIcon()

    @staticmethod
    def workflow_constructor(*args, **kwargs):
        return WidgetsScheme(*args, **kwargs)

    @staticmethod
    def widgets_entry_points():
        """
        Return an `EntryPoint` iterator for all 'orange.widget' entry
        points.
        """
        # Ensure the 'this' distribution's ep is the first. iter_entry_points
        # yields them in unspecified order.
        WIDGETS_ENTRY = "orange.widgets"

        def is_tomwer_extension(entry):
            return "tomwer" in entry.name.lower()

        all_eps = filter(
            is_tomwer_extension, pkg_resources.iter_entry_points(WIDGETS_ENTRY)
        )

        all_eps = sorted(
            all_eps,
            key=lambda ep: 0
            if ep.dist.project_name.lower() in ("orange3", "tomwer")
            else 1,
        )
        return iter(all_eps)

    @staticmethod
    def addon_entry_points():
        return TomwerConfig.widgets_entry_points()

    APPLICATION_URLS = {
        #: Submit a bug report action in the Help menu
        "Bug Report": "https://gitlab.esrf.fr/tomotools/tomwer/-/issues",
        #: A url quick tour/getting started url
        "Quick Start": "https://www.youtube.com/playlist?list=PLddRXwP6Z6F9KOu1V5o6H24KPH5Ikuk2f",
        #: The 'full' documentation, should be something like current /docs/
        #: but specific for 'Visual Programing' only
        "Documentation": "https://tomotools.gitlab-pages.esrf.fr/nxtomomill/index.html",
        #: YouTube tutorials
        "Screencasts": "https://www.youtube.com/@tomotools",
    }


class TomwerSplashScreen(qt.QSplashScreen):
    def __init__(
        self,
        parent=None,
        pixmap=None,
        textRect=None,
        textFormat=qt.Qt.PlainText,
        **kwargs,
    ):
        super(TomwerSplashScreen, self).__init__(pixmap=icons.getQPixmap("tomwer"))

    def showMessage(self, message, alignment=qt.Qt.AlignLeft, color=qt.Qt.black):
        version = f"tomwer version {tomwer.version.version}"
        super().showMessage(version, qt.Qt.AlignLeft | qt.Qt.AlignBottom, qt.Qt.white)


class TerminalTextDocument(_TerminalTextDocument):
    def __init__(self, parent=None, **kwargs):
        super().__init__(parent=parent, **kwargs)

    @staticmethod
    def get_log_level(my_str):
        if "DEBUG" in my_str:
            return qt.Qt.darkBlue
        elif "ERROR" in my_str:
            return qt.Qt.red
        elif "WARNING" in my_str:
            return qt.Qt.magenta
        elif "CRITICAL" in my_str:
            return qt.Qt.red
        elif "Info" in my_str:
            return qt.Qt.black
        elif "CRITICAL" in my_str:
            return qt.Qt.darkYellow
        elif "PROCESS_STARTED" in my_str:
            return qt.Qt.black
        elif "PROCESS_SUCCEED" in my_str:
            return qt.Qt.darkGreen
        elif "PROCESS_FAILED" in my_str:
            return qt.Qt.red
        elif "PROCESS_ENDED" in my_str:
            return qt.Qt.black
        elif "PROCESS_SKIPPED" in my_str:
            return qt.Qt.magenta
        else:
            return None

    def writeWithFormat(self, string: str, charformat) -> None:
        assert qt.QThread.currentThread() is self.thread()
        # remove linux reset sequence
        string = string.replace("\033[0m", "")
        # remove linux reset sequence
        string = string.replace("\033[1;%dm", "")
        # remove linux reset sequence
        string = string.replace("\x1B[1m", "")
        string = string.replace("\x1B[1;30m", "")
        string = string.replace("\x1B[1;31m", "")
        string = string.replace("\x1B[1;32m", "")
        string = string.replace("\x1B[1;33m", "")
        string = string.replace("\x1B[1;34m", "")
        string = string.replace("\x1B[1;35m", "")

        color = self.get_log_level(string) or qt.Qt.red
        charformat.setForeground(color)

        # super().writelinesWithFormat(string, charformat)
        cursor = self.textCursor()
        cursor.setCharFormat(charformat)
        cursor.insertText(string)



from ewokscore.task import Task as EwoksTask


class _TomoobjseriePlaceHolder(
    EwoksTask, input_names=["serie"], output_names=["serie"]
):
    def run(self):
        self.outputs.serie = self.inputs.serie



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017-2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "10/12/2021"


import gc
import pytest
from silx.gui import qt
from tomwer.tests.utils import skip_gui_test
from silx.gui.utils.testutils import TestCaseQt
from orangecontrib.tomwer.widgets.edit.ImageKeyEditorOW import ImageKeyEditorOW


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestImageKeyEditor(TestCaseQt):
    """Test that the ImageKeyEditorOW widget work correctly (at least launched)"""

    def setUp(self):
        super().setUp()
        self._window = ImageKeyEditorOW()

    def tearDown(self):
        self._window.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._window.close()
        self._window = None
        gc.collect()

    def test(self):
        self._window.show()
        self.qWaitForWindowExposed(self._window)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "18/02/2021"


from silx.utils.enum import Enum as _Enum
import numpy
import sys
import logging
from typing import Union

_logger = logging.getLogger(__name__)


class ScoreMethod(_Enum):
    STD = "standard deviation"
    TV = "total variation"
    TV_INVERSE = "1 / (total variation)"
    STD_INVERSE = "1 / std"
    TOMO_CONSISTENCY = "tomo consistency"


class ComputedScore:
    def __init__(self, tv, std, tomo_consistency=None):
        self._tv = tv
        self._std = std
        self._tomo_consistency = tomo_consistency

    @property
    def total_variation(self):
        return self._tv

    @property
    def std(self):
        return self._std

    @property
    def tomo_consistency(self):
        return self._tomo_consistency

    def get(self, method: ScoreMethod):
        method = ScoreMethod.from_value(method)
        if method is ScoreMethod.TV:
            return self.total_variation / float(10e5)
        elif method is ScoreMethod.TV_INVERSE:
            return (1.0 / self.total_variation) * float(10e5)
        elif method is ScoreMethod.STD:
            return self.std
        elif method is ScoreMethod.STD_INVERSE:
            return 1.0 / self.std
        elif method is ScoreMethod.TOMO_CONSISTENCY:
            return self.tomo_consistency
        else:
            raise ValueError("{} is an unrecognized method".format(method))


def compute_tomo_consistency(
    data: numpy.ndarray,
    original_sinogram,
    angles,
    original_axis_position,
    detector_width,
):
    from silx.opencl.projection import Projection

    projector = Projection(
        slice_shape=data.shape,
        angles=angles,
        detector_width=detector_width,
        axis_position=original_axis_position,
    )
    sinogram = projector.projection(data)
    sinogram_n = (sinogram - sinogram.min()) / (sinogram.max() - sinogram.min())
    original_sinogram_n = (original_sinogram - original_sinogram.min()) / (
        original_sinogram.max() - original_sinogram.min()
    )
    return 1.0 / (numpy.sum(numpy.abs(sinogram_n - original_sinogram_n)) + 1)


def compute_score_contrast_std(data: numpy.ndarray):
    """
    Compute a contrast score by simply computing the standard deviation of
    the frame
    :param numpy.ndarray data: frame for which we should compute the score
    :return: score of the frame
    :rtype: float
    """
    if data is None:
        return None
    else:
        return data.std() * 100


def compute_tv_score(data: numpy.ndarray):
    """
    Compute the data score as image total variation

    :param numpy.ndarray data: frame for which we should compute the score
    :return: score of the frame
    :rtype: float
    """
    tv = numpy.sum(
        numpy.sqrt(
            numpy.gradient(data, axis=0) ** 2 + numpy.gradient(data, axis=1) ** 2
        )
    )
    return tv


_METHOD_TO_FCT = {
    ScoreMethod.STD: compute_score_contrast_std,
    ScoreMethod.TV: compute_tv_score,
}


def compute_score(
    data: numpy.ndarray,
    method: ScoreMethod,
    angles: Union[None, list] = None,
    original_sinogram: Union[None, numpy.array] = None,
    original_axis_position: Union[float, None] = None,
    detector_width: Union[float, None] = None,
) -> Union[float, None]:
    """

    :param numpy.ndarray data: frame for which we should compute the score
    :param str method:
    :return: score of the frame
    :rtype: float
    """
    method = ScoreMethod.from_value(method)
    if data.ndim == 3:
        if data.shape[0] == 1:
            data = data.reshape(data.shape[1], data.shape[2])
        elif data.shape[2] == 1:
            data = data.reshape(data.shape[0], data.shape[1])
        else:
            raise ValueError("Data is expected to be 2D. Not {}".format(data.ndim))
    elif data.ndim == 2:
        pass
    else:
        raise ValueError("Data is expected to be 2D. Not {}".format(data.ndim))
    if method is ScoreMethod.TOMO_CONSISTENCY:
        try:
            return compute_tomo_consistency(
                data=data,
                angles=angles,
                original_sinogram=original_sinogram,
                original_axis_position=original_axis_position,
                detector_width=detector_width,
            )
        except ImportError:
            return None
        except Exception as e:
            _logger.warning(
                "Fail to compute 'tomo consistency' score. " "Reason is {}".format(e)
            )
            return None
    else:
        fct = _METHOD_TO_FCT.get(method, None)

    if fct is not None:
        return fct(data)
    else:
        raise ValueError("{} is not handled".format(method))


def get_disk_mask_radius(datasets_) -> int:
    """compute the radius to use for the mask"""
    radius = sys.maxsize
    # get min radius
    for _, (_, data) in datasets_.items():
        if data is None:
            continue
        assert data.ndim == 2, "data is expected to be 2D"
        min_ = numpy.array(data.shape).min()
        if radius >= min_:
            radius = min_
    return radius // 2


def apply_roi(data, radius, url) -> numpy.array:
    """compute the square included in the circle of radius and centered
    in the middle of the data"""
    half_width = int(radius / 2**0.5)
    center = numpy.array(data.shape[:]) // 2
    min_x, max_x = center[0] - half_width, center[0] + half_width
    min_y, max_y = center[1] - half_width, center[1] + half_width
    try:
        return data[min_y:max_y, min_x:max_x]
    except Exception:
        _logger.error(
            "Fail to apply roi for {}. Take the entire dataset".format(url.path())
        )
        return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import argparse
from silx.gui import qt
import signal
import functools
from tomwer.gui import icons
from tomwer.gui.utils.splashscreen import getMainSplashScreen
from tomwer.gui.reconstruction.axis import AxisWindow
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.synctools.axis import QAxisRP
from tomwer.core.process.reconstruction.axis.axis import AxisProcess, NoAxisUrl
from tomwer.core.process.reconstruction.axis.axis import AxisMode
from tomwer.core.process.reconstruction.darkref.darkrefs import (
    requires_reduced_dark_and_flat,
)
from tomwer.core.utils.resource import increase_max_number_file
from silx.io.url import DataUrl
import logging


logging.basicConfig(level=logging.WARNING)
_logger = logging.getLogger(__name__)


class _AxisProcessGUI(AxisWindow):
    def __init__(self, scan, axis_params):
        axis_params.frame_width = scan.dim_1
        AxisWindow.__init__(self, axis_params=axis_params)
        self.hideLockButton()
        self.hideApplyButton()
        self.setScan(scan=scan)

        # connect Signal / Slot
        callback = functools.partial(self.compute, scan)
        self.sigComputationRequested.connect(callback)
        self.setWindowIcon(icons.getQIcon("tomwer"))

    def compute(self, scan, wait=True):
        mess = " ".join(
            ("start axis calculation with", scan.axis_params.axis_url_1.url.path())
        )
        _logger.info(mess)
        process = AxisProcess(
            inputs={
                "data": scan,
                "axis_params": self._axis_params,
                "wait": wait,
            }
        )
        try:
            process.run()
        except NoAxisUrl:
            msg = qt.QMessageBox(self)
            msg.setIcon(qt.QMessageBox.Warning)
            text = (
                "Unable to find url to compute the axis, please select them "
                "from the `axis input` tab"
            )
            msg.setText(text)
            msg.exec_()
            return None
        else:
            position = scan.axis_params.relative_cor_value
            assert isinstance(position, (float, type(None)))
            frm = "sinogram" if scan.axis_params.use_sinogram else "radio"
            AxisWindow.setPosition(self, frm=frm, value=position)
            return position

    def _updatePosition(self, scan):
        self._widget.setPosition(frm=None, value=scan.axis_params.relative_cor_value)


def main(argv):
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "--scan-path",
        help="For EDF acquisition: provide folder path, for HDF5 / nexus "
        "provide the master file",
        default=None,
    )
    parser.add_argument(
        "--use-sinogram",
        help="use the signoram from radio for computing COR. This only work"
        "with the =scan-path option",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--radio1",
        help="If you cannot provide scan-path or master file you should "
        "provide urls of the two radios to use for calculation",
        default=None,
    )
    parser.add_argument(
        "--radio2",
        help="If you cannot provide scan-path or master file you should "
        "provide urls of the two radios to use for calculation",
        default=None,
    )
    parser.add_argument(
        "--entry", help="For Nexus files: entry in the master file", default=None
    )
    parser.add_argument(
        "--debug",
        dest="debug",
        action="store_true",
        default=False,
        help="Set logging system in debug mode",
    )
    parser.add_argument(
        "--mode",
        default=None,
        help="Use a specific mode. Available modes are {}" "".format(AxisMode.values()),
    )
    parser.add_argument(
        "--full-image",
        action="store_true",
        default=False,
        help="Define the mode 'full' when display shifted images from manual" " mode",
    )

    options = parser.parse_args(argv[1:])

    if options.debug:
        logging.root.setLevel(logging.DEBUG)

    increase_max_number_file()

    global app  # QApplication must be global to avoid seg fault on quit
    app = qt.QApplication.instance() or qt.QApplication([])
    splash = getMainSplashScreen()
    qt.QApplication.setOverrideCursor(qt.Qt.WaitCursor)
    qt.QApplication.processEvents()

    qt.QLocale.setDefault(qt.QLocale(qt.QLocale.English))
    qt.QLocale.setDefault(qt.QLocale.c())
    signal.signal(signal.SIGINT, sigintHandler)
    sys.excepthook = qt.exceptionHandler

    timer = qt.QTimer()
    timer.start(500)
    # Application have to wake up Python interpreter, else SIGINT is not
    # catch
    timer.timeout.connect(lambda: None)

    mode = options.mode
    if mode is not None:
        mode = AxisMode.from_value(mode)

    try:
        scan = ScanFactory.create_scan_object(
            scan_path=options.scan_path, entry=options.entry
        )
    except ValueError:
        scan = ScanFactory.mock_scan()

    requires_reduced_dark_and_flat(scan=scan, logger_=_logger)
    # define the process_index is any tomwer_processes_existing
    scan.set_process_index_frm_tomwer_process_file()
    if scan.axis_params is None:
        scan.axis_params = QAxisRP()
    if options.radio1 is not None:
        radio1 = DataUrl(file_path=options.radio1, scheme="fabio")
        scan.axis_params.axis_url_1 = radio1
        if options.radio2 is not None:
            radio2 = DataUrl(file_path=options.radio2, scheme="fabio")
        else:
            radio2 = None
        scan.axis_params.axis_url_2 = radio2

    if options.debug:
        _logger.setLevel(logging.DEBUG)

    axis_params = QAxisRP()
    axis_params.use_sinogram = options.use_sinogram
    if scan.dim_2 is not None:
        axis_params.sinogram_line = scan.dim_2 // 2

    window = _AxisProcessGUI(scan=scan, axis_params=axis_params)
    window.setWindowTitle("axis")
    window.setWindowIcon(icons.getQIcon("tomwer"))
    if mode is not None:
        window.setMode(mode)
    if options.full_image:
        window.manual_uses_full_image(True)

    splash.finish(window)
    window.show()
    qt.QApplication.restoreOverrideCursor()
    app.exec_()


def getinputinfo():
    return "tomwer axis [scanDir]"


def sigintHandler(*args):
    """Handler for the SIGINT signal."""
    qt.QApplication.quit()


if __name__ == "__main__":
    main(sys.argv)
